#
# Author: Ben Dudson, Department of Physics, University of York
#         benjamin.dudson@york.ac.uk
#
# This file is part of PyXPad.
#
# PyXPad is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyXPad is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

from numpy import sqrt, abs, max


class XPadDataDim:
    """
    Dimension of a data item

    name     Short name (e.g. "t")
    label    Short axis label (e.g. "Time (sec)")
    units    (e.g. "s")
    data     Axis values (NumPy array)
    errl     Low-side error (may be None)
    errh     High-side error (may be None)
    """

    def __init__(self, other=None):  # Constructor
        # Instance Variables
        self.name = ""
        self.label = ""
        self.units = ""
        self.data = None
        self.errl = None
        self.errh = None

        if other is not None:
            try:
                # List of variables to copy
                varlist = ["name", "label", "units",
                           "data", "errl", "errh"]
                for name in varlist:
                    # Check if other has this property
                    try:
                        setattr(self, name, getattr(other, name))
                    except AttributeError:
                        pass
                if self.name == "":
                    self.name = self.label
            except:
                pass

    def __str__(self):
        return self.name

    def __repr__(self):
        return ("XPadDataDim( {'name':'"+self.name +
                "', 'label':'"+self.label +
                "', 'units':'"+self.units+"'} )")

    def __eq__(self, other):
        return ((self.name == other.name) and
                (self.units == other.units) and
                all(self.data == other.data))


class XPadDataItem:
    """
    Data item class for PyXPad. Provides a standard interface
    and numerical operators

    name    The name used to request the data (e.g. "amc_plasma current")
    source  Source of the data as a string (e.g. "15100")
    label   Short description (e.g. "Plasma Current")
    units   Data units (e.g. "kA")
    desc    longer description (if set)
    data    NumPy array of the data
    errl    Low-side error (may be None)
    errh    High-side error (may be None)
    dim     A list of dimensions, each of which contains:
      - label  Short axis label (e.g. "Time (sec)")
      - units  (e.g. "s")
      - data   Axis values (NumPy array)
      - errl   Low-side error (may be None)
      - errh   High-side error (may be None)
    order   Index of time dimension
    time    A shortcut to the time data (dim[order].data). May be None

    """

    def __init__(self, other=None):  # Constructor
        # Instance Variables
        self.name   = ""
        self.source = ""
        self.label  = ""
        self.units  = ""
        self.desc   = ""
        self.data   = None
        self.errl   = None
        self.errh   = None
        self.rank   = None
        self.dim    = XPadDataDim()             # A list of dimensions
        self.order  = -1             # Index of time dimension
        self.time   = None           # A shortcut to the time data (dim[order].data). May be None

        if other is not None:
            if hasattr(other, "data"):
                # List of variables to copy
                varlist = ["name", "source", "label", "units", "desc",
                           "data", "rank", "errl", "errh", "order", "time"]
                for name in varlist:
                    # Check if other has this property
                    try:
                        setattr(self, name, getattr(other, name))
                    except AttributeError:
                        pass
                if self.name == "":
                    self.name = other.label
                try:
                    self.dim = [XPadDataDim(dim) for
                                dim in other.dim]
                except AttributeError:
                    pass
            else:
                # Assume it's a numerical type
                self.data = other
                self.name = str(other)

    # def __coerce__(self, other):
    #    # Convert other to an XPadDataItem and return
    #    item = XPadDataItem

    def __str__(self):
        """
        Returns a summary of the data as a string
        """
        s = self.name + "("+self.units+")"

        if len(self.dim) > 0:
            s += " [" + ",".join([str(d) for d in self.dim]) + "]"

        return s

    def __repr__(self):
        return ("XPadDataItem( {'name':'"+self.name +
                "', 'source':'"+self.source +
                "', 'label':'"+self.label +
                "', 'units':'"+self.units +
                "', 'desc':'"+self.desc+"'} )")

    def __add__(self, other):  # +
        item = XPadDataItem(self)
        item += other
        return item

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):  # +=
        try:
            # Metadata
            self.name += " + " + other.name
            if (self.label != "") and (other.label != ""):
                self.label += " + " + other.label
            else:
                self.label = self.name

            # Dimensions
            if (self.dim == other.dim):
                self.dim = other.dim
            else:
                raise ValueError("Incompatible dims: {} and {}".format(
                    self.dim, other.dim))

            # Low-side error
            if self.errl is not None and other.errl is not None:
                self.errl = sqrt(self.errl**2 + other.errl**2)
            elif other.errl is not None:
                self.errl = other.errl

            # High-side error
            if self.errh is not None and other.errh is not None:
                self.errh = sqrt(self.errh**2 + other.errh**2)
            elif other.errh is not None:
                self.errh = other.errh

            # Data
            self.data = self.data + other.data
        except AttributeError:
            # other probably just a numeric type
            self.name += " + " + str(other)
            if self.label != "":
                self.label += " + " + str(other)
            self.data = self.data + other

        return self

    def __sub__(self, other):  # -
        item = XPadDataItem(self)
        item -= other
        return item

    def __rsub__(self, other):  # -
        item = -(self - other)  # Lazy way
        return item

    def __isub__(self, other):         # -=
        try:
            # Metadata
            self.name += " - " + other.name
            if (self.label != "") and (other.label != ""):
                self.label += " - " + other.label
            else:
                self.label = self.name

            # Dimensions
            if (self.dim == other.dim):
                self.dim = other.dim
            else:
                raise ValueError("Incompatible dims: {} and {}".format(
                    self.dim, other.dim))

            # Low-side error. Note h and l swap for other
            if self.errl is not None and other.errh is not None:
                self.errl = sqrt(self.errl**2 + other.errh**2)
            elif other.errh is not None:
                self.errl = other.errh

            # High-side error
            if self.errh is not None and other.errl is not None:
                self.errh = sqrt(self.errh**2 + other.errl**2)
            elif other.errl is not None:
                self.errh = other.errl

            # Data
            self.data = self.data - other.data
        except:
            self.name += " - " + str(other)
            if self.label != "":
                self.label += " - " + str(other)
            self.data = self.data - other
        return self

    def __mul__(self, other):  # *
        item = XPadDataItem(self)
        item *= other
        return item

    def __rmul__(self, other):
        return self.__mul__(other)

    def __imul__(self, other):         # *=
        try:
            # Metadata
            self.name += " * " + other.name
            if (self.label != "") and (other.label != ""):
                self.label += " * " + other.label
            else:
                self.label = self.name

            # Dimensions
            if (self.dim == other.dim):
                self.dim = other.dim
            else:
                raise ValueError("Incompatible dims: {} and {}".format(
                    self.dim, other.dim))

            # Units
            if self.units == other.units:
                self.units += chr(0x00B2)
            else:
                self.units += other.units

            # Low-side error
            if self.errl is not None and other.errl is not None:
                self.errl = sqrt( (other.data*self.errl)**2 + (self.data * other.errl)**2 )
            elif other.errl is not None:
                self.errl = self.data * other.errl
            elif self.errl is not None:
                self.errl = other.data * self.errl

            # High-side error
            if self.errh is not None and other.errh is not None:
                self.errh = sqrt( (other.data*self.errh)**2 + (self.data * other.errh)**2 )
            elif other.errh is not None:
                self.errh = self.data * other.errh
            elif self.errh is not None:
                self.errh = other.data * self.errh

            # Data
            self.data = self.data * other.data
        except:
            self.name = "( " + self.name + " * " + str(other) + " )"
            if self.label != "":
                self.label = "( " + self.label + " * " + str(other) + " )"
            self.data = self.data * other
            if self.errl is not None:
                self.errl = self.errl * other
            if self.errh is not None:
                self.errh = self.errh * other
        return self

    def __truediv__(self, other):  # /
        item = XPadDataItem(self)
        item /= other
        return item

    def __itruediv__(self, other):  # /=
        try:
            # Metadata
            self.name += " / " + other.name
            if (self.label != "") and (other.label != ""):
                self.label += " / " + other.label
            else:
                self.label = self.name

            # Dimensions
            if (self.dim == other.dim):
                self.dim = other.dim
            else:
                raise ValueError("Incompatible dims: {} and {}".format(
                    self.dim, other.dim))

            #Units
            if self.units == other.units:
                self.units = ""
            elif self.units == "":
                self.units = other.units + chr(0x207B) + chr(0x00B9)
            else:
                self.units += "/" + other.units

            # Low-side error. Note h and l swap for other
            if self.errl is not None and other.errh is not None:
                self.errl = sqrt((self.errl / other.data)**2 + (self.data * other.errh / other.data**2)**2)
            elif other.errh is not None:
                self.errl = self.data * other.errh / other.data**2
            elif self.errl is not None:
                self.errl = self.errl / other.data

            # High-side error
            if self.errh is not None and other.errl is not None:
                self.errh = sqrt((self.errh / other.data)**2 + (self.data * other.errl / other.data**2)**2)
            elif other.errl is not None:
                self.errh = self.data * other.errl / other.data**2
            elif self.errh is not None:
                self.errh = self.errh / other.data

            # Data
            self.data = self.data / other.data
        except AttributeError:
            self.name = "( " + self.name + " / " + str(other)+" )"
            if self.label != "":
                self.label = "( " + self.label + " / " + str(other)+" )"
            self.data = self.data / other
            if self.errl is not None:
                self.errl = self.errl / other
            if self.errh is not None:
                self.errh = self.errh / other
        return self

    def __rtruediv__(self, other):  #
        item = XPadDataItem(other)
        item /= self
        return item

    def __neg__(self):  # Unary minus
        item = XPadDataItem(self)
        item.name = "-"+self.name
        if self.label != "":
            item.label = "-"+self.label

        item.data = -self.data
        # Swap high and low errors
        item.errl = self.errh
        item.errh = self.errl

        return item

    def __pos__(self):
        return self

    def __abs__(self):
        item = XPadDataItem(self)
        item.name = "abs( "+self.name+" )"
        if self.label != "":
            item.label = "abs( "+self.label+" )"

        item.data = abs(self.data)
        # High side error is maximum of low and high
        if self.errl is not None and self.errh is not None:
            pass
        if self.errl is not None:
            item.errh = self.errl

        # Low side error is zero
        item.errl = 0.0

        return item


def chop(item):
    """
    Selects a range of indices

    """
    pass

if __name__ == "__main__":
    # Run test cases

    a = XPadDataItem()
    a.name = "a"
    a.data = 1
    a.errl = 0.1
    a.errh = 0.2

    b = abs(a*3 + 2)

    c = 2 * a

    d = 4 / a
    print(d.data, d.name)

    print(b.data, b.errl, b.errh)
    print(b.name)



# A wrapper around various NetCDF libraries,
#
# Supported libraries:
# -------------------
#
# netCDF4
#
# Scientific.IO.NetCDF
#
# scipy.io.netcdf
#   old version (create_dimension, create_variable)
#   new version (createDimension, createVariable)
#

try:
    import numpy as np
except ImportError:
    print("ERROR: NumPy module not available")
    raise

library = None  # Record which library to use

try:
    from netCDF4 import Dataset
    library = "netCDF4"
except ImportError:
    #print("netcdf4-python module not found")

    try:
        from Scientific.IO.NetCDF import NetCDFFile as Dataset
        from Scientific.N import Int, Float
        library = "Scientific"
        #print("  => Using Scientific.IO.NetCDF instead")
    except ImportError:
        try:
            from scipy.io.netcdf import netcdf_file as Dataset
            library = "scipy"
            #print("Using scipy.io.netcdf library")
        except:
            print("No supported NetCDF modules available")
            raise
import time

from .pyxpad_utils import XPadDataItem, XPadDataDim


class NetCDFDataSource:
    """

    Functions
      read( name , shot)   Input variable name (string)
                           Output is an XPadDataItem object or None

      size( name )   Returns variable size as a list. [] for scalar

    Attributes

      label         A short string to describe the source
      dimensions    A dictionary of XPadDataDim objects
      varNames      A list of variable names
      variables     A dictionary of XPadDataItem objects with empty data

    """
    handle = None

    def open(self, fname=None):
        if fname is None:
            fname = self.filename
        self.handle = Dataset(fname, "r")

    def close(self):
        if self.handle is not None:
            self.handle.close()
        self.handle = None

    def __init__(self, filename):
        self.filename = filename
        self.label = filename   # May need to shorten
        self.open(filename)
        self.dimensions = self.getDimensions()        # A dictionary of XPadDataDim objects
        self.varNames = self.handle.variables.keys()  # A list of variable names
        for i, v in enumerate(self.varNames):
            try:
                # Python 2
                if isinstance(v, unicode):
                    v = v.encode('utf-8')
                v = str(v).translate(None, '\0')
            except NameError:
                # Python 3
                if isinstance(v, str):
                    v = v.encode('utf-8')

            self.varNames[i] = v

        self.variables = {}   # A dictionary of XPadDataItem objects with empty data
        for name, var in self.handle.variables.items():
            item = XPadDataItem()
            item.name   = name
            item.source = self.filename
            item.dim = map(lambda d: self.dimensions[d], var.dimensions)
            self.variables[name] = item
        self.close()

    def __del__(self):
        self.close()

    def read(self, name, shot):
        """Read a variable from the file."""
        self.open()
        if self.handle is None:
            return None

        try:
            var = self.handle.variables[name]
        except KeyError:
            # Not found. Try to find using case-insensitive search
            var = None
            for n in self.handle.variables.keys():
                if n.lower() == name.lower():
                    print("WARNING: Reading '"+n+"' instead of '"+name+"'")
                    var = self.handle.variables[n]
            if var is None:
                return None
        ndims = len(var.dimensions)
        if ndims == 0:
            data = var.getValue()
        else:
            data = var[:]
        item = XPadDataItem()
        item.name   = name
        item.source = self.filename
        item.data   = data
        item.dim = map(lambda d: self.dimensions[d], var.dimensions)

        self.close()
        return item

    def getDimensions(self):
        if self.handle is None:
            return None
        dims = {}
        for name, dim in self.handle.dimensions.items():
            t = type(dim).__name__
            if t == 'int':
                n = dim
            else:
                n = len(dim)
            newdim = XPadDataDim()
            newdim.name = name
            newdim.label = name
            newdim.data  = np.arange(n)
            dims[name] = newdim
        return dims

    def size(self, varname):
        """List of dimension sizes for a variable."""
        if self.handle is None:
            return []
        try:
            var = self.handle.variables[varname]
        except KeyError:
            return []

        def dimlen(d):
            dim = self.handle.dimensions[d]
            if dim is not None:
                t = type(dim).__name__
                if t == 'int':
                    return dim
                return len(dim)
            return 0
        return map(lambda d: dimlen(d), var.dimensions)



"""
Functions which can be used to manipulate data items

Available to the user in the command terminal

"""

import numpy as np
from pyxpad import calculus
from .pyxpad_utils import XPadDataItem, XPadDataDim


def XPadFunction(func, name="f"):
    """
    Turns a NumPy function into a function of data item
    """
    def newfunc(data):
        result = XPadDataItem()
        if data.name != "":
            result.name = name + "( " + data.name + " )"
        result.source = data.source
        if data.label != "":
            result.label = name + "( " + data.label + " )"
        result.data   = func(data.data)
        result.dim    = data.dim
        result.order  = data.order
        result.time   = data.time
        return result
    return newfunc


sin = XPadFunction(np.sin, "sin")
cos = XPadFunction(np.cos, "cos")
tan = XPadFunction(np.tan, "tan")

exp = XPadFunction(np.exp, "exp")
log = XPadFunction(np.log, "log")

sqrt = XPadFunction(np.sqrt, "sqrt")

def reciprocal(data):
    recip = XPadFunction(np.reciprocal, "recip")
    reciprocal = recip(data)
    reciprocal.units = data.units + chr(0x207B) + chr(0x00B9)
    return reciprocal

def exponential(data):
    exp = XPadFunction(np.exp, "exp")
    return exp(data)

def absolute(data):
    abso = XPadFunction(np.absolute, "abs")
    return abso(data)

def arctan(data):
    atan = XPadFunction(np.arctan, "arctan")
    return atan(data)

def nlog(data):
    natlog = XPadFunction(np.log, "nlog")
    return natlog(data)

def normalise(data):
    point = len(data.data)-1
    integral = calculus.integrate(data)
    normfac = integral.data[point]
    normdat = np.true_divide(data.data, normfac)
    result = XPadDataItem(integral)
    result.data[:] = normdat[:]
    if data.name != "":
        result.name = "Norm(" + data.name + ")"
    if data.label != "":
        result.label = "Norm(" + data.label + ")"
    result.units = data.units
    return result

def invert(data):
    result = XPadDataItem(data)
    result.data = -1.*data.data
    result.name = "-(" + data.name + ")"
    result.label = "-(" + data.label + ")"
    return result

from PySide.QtGui import QDialog, QGridLayout, QDialogButtonBox, QPushButton
from PySide.QtCore import Qt
from PySide.QtGui import QInputDialog

def addcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X+C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    result.data = np.add(data.data, constant)
    result.name = "(" + data.name + ")+" + str(constant)
    result.label = "(" + data.label + ")+" + str(constant)
    return result

def subcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X-C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    constant = 1.
    result.data = np.subtract(data.data, constant)
    result.name = "(" + data.name + ")-" + str(constant)
    result.label = "(" + data.label + ")-" + str(constant)
    return result

def mulcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X*C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    result.data = np.multiply(data.data, constant)
    result.name = "(" + data.name + ")*" + str(constant)
    result.label = "(" + data.label + ")*" + str(constant)
    return result

def divcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X/C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    result.data = np.true_divide(data.data, constant)
    result.name = "(" + data.name + ")/" + str(constant)
    result.label = "(" + data.label + ")/" + str(constant)
    return result

def powcon(data):
    result = XPadDataItem(data)
    parent = QDialog()
    dialog = QInputDialog.getDouble(parent, "X^C", "Input constant:")
    if dialog[1]:
        constant = dialog[0]
    result.data = np.power(data.data, constant)
    result.name = "(" + data.name + ")^" + str(constant)
    result.label = "(" + data.label + ")^" + str(constant)
    result.units = "(" + data.units + ")^" + str(constant)
    return result

def inputname():
    parent = QDialog()
    title = "Change Name"
    label = "Enter new name:"
    dialog = QInputDialog.getText(parent, title, label)
    if dialog[1]:
        name = dialog[0]
    return name

def changename(data):
    result = XPadDataItem(data)
    result.name = "Renamed(" + data.name +")"
    return result

def changeunits(data):
    result = XPadDataItem(data)
    parent = QDialog()
    title = "Change Units"
    label = "Enter new units:"
    dialog = QInputDialog.getText(parent, title, label)
    if dialog[1]:
        newUnit = dialog[0]
        unitFact = getUnitFactor()
    if unitFact and newUnit:
        result.data = np.multiply(data.data, unitFact)
        result.units = newUnit
        return result

def getUnitFactor():
    parent = QDialog()
    title = "Change Units"
    label = "Enter conversion factor:"
    dialog = QInputDialog.getDouble(parent, title, label)
    if dialog[1]:
        return dialog[0]
    else:
        return None

def statistics(data):
    n = len(data.data)
    sumOfData = np.sum(data.data)
    mean = sumOfData/n
    dataSquared = np.power(data.data, 2)
    variance = (np.sum(dataSquared)/n)-mean**2.
    stdDev = np.sqrt(variance)
    minVal = data.data[0]
    maxVal = data.data[0]
    for value in data.data:
        if value >= maxVal:
            maxVal = value
        if value <= minVal:
            minVal = value
    rng = "(" + str(minVal) +")" + " - " + "(" + str(maxVal) + ")"
    stats = "\nStatistics on " + data.name + ":"
    stats += "\nMean = " + str(mean) + "\nStandard Deviation = " + str(stdDev)
    stats += "\nRange = " + rng
    print(stats)
    return()

def timeOffset(data):
    result = XPadDataItem(data)
    parent = QDialog()
    title = "Time Offset"
    label = "Input time offest:"
    dialog = QInputDialog.getDouble(parent, title, label)
    if dialog[1]:
        offset = dialog[0]
        result.dim[data.order].data = np.add(result.dim[data.order].data, offset)
        result.name = "Timoff(" + result.name + ", " + str(offset) + ")"
        result.label = "Timoff(" + result.label + ", " + str(offset) + ")"
        return result

def chop(item, t_min, t_max):
    """
        >>> from user_functions import *
        >>> a = chop(XMC_OMV_110, 0.274, 0.276)
        >>> a_amp,a_phase = fftp(a)
        >>> b = chop(a_phase, 0.0, 100.0)
        >>> plot(b)
    """
    if len(item.dim) != 1:
        raise ValueError("chop can only operate on 1D traces currently")

    if t_max < t_min or t_max < item.dim[0].data[0] or t_min > item.dim[0].data[-1]:
        raise ValueError("New time-range not defined correctly")

    idx = np.where(np.logical_and(item.dim[0].data >= t_min, item.dim[0].data <= t_max))

    if len(idx[0]) == 0:
        raise ValueError("No data in time-range specified")

    # Calculate the phase
    chopped = XPadDataItem()

    if item.name != "":
        chopped.name = "CHOP( "+item.name+", "+str(t_min)+", "+str(t_max)+" )"
    chopped.source = item.source
    if item.label != "":
        chopped.label = "CHOP( "+item.label+", "+str(t_min)+", "+str(t_max)+" )"
    chopped.units = item.units

    chopped.data = item.data[idx]

    # Create a dimension
    dim = XPadDataDim()

    dim.name = item.dim[0].name
    dim.label = item.dim[0].label
    dim.units = item.dim[0].units

    dim.data = item.dim[0].data[idx]
    chopped.dim = [dim]

    if chopped.dim[0].units in ["s", "S", "sec", "Sec", "SEC"]:
        chopped.order = 0
        chopped.time = chopped.dim[0].data

    return chopped

def clip(item, valmin, valmax):
    """
    Removes values outside a given range
    """

    if len(item.dim) != 1:
        raise ValueError("Clip can only operate on 1D traces currently")

    if valmax < valmin:
        raise ValueError("Clip range incorrectly defined")

    clipped = XPadDataItem(item)
    data = []
    time = []

    for point in range(len(item.data)-1):
        if item.data[point] <= valmax and item.data[point] >= valmin:
            data += [item.data[point]]
            time += [item.dim[item.order].data[point]]

    if len(data) == 0:
        raise ValueError("No data in the specified range")

    clipped.data = data
    clipped.dim[clipped.order].data = time
    clipped.time = time
    clipped.name = "CLIP("+item.name+", "+str(valmin)+", "+str(valmax)+")"
    clipped.label = "CLIP("+item.label+", "+str(valmin)+", "+str(valmax)+")"

    return clipped

##########################################################################

"""
    from user_functions import *
    data = read_padsav("12354-omv110-FFTP.padsav")[0]
    plot(data)
"""

from numpy import linspace, rollaxis
from scipy.io import readsav
from warnings import catch_warnings, simplefilter


def read_padsav(file_name, disable_UserWarnings=True):
    """
        Reads data from an idl_dict object into XPadDataItem objects

        Parameters
        ----------
        file_name : str
            Path to XPad*.padsav file


        Returns
        -------
        items : list
            A list of XPadDataItems
    """

    warning_action = "default"
    if (disable_UserWarnings):
        warning_action = "ignore"

    with catch_warnings():
        simplefilter(warning_action, UserWarning)
        idl_dict = readsav(file_name)

    return parse_padsav(idl_dict)


def parse_padsav(xpad_idl_dict):
    """
        Reads data from an idl_dict object into XPadDataItem objects

        Parameters
        ----------
        xpad_idl_dict : AttrDict or dict
            Input dict from an XPad *.padsav imported using scipy.io.readsav


        Returns
        -------
        items : list
            A list of XPadDataItems
    """

    if not check_padsav(xpad_idl_dict):
        return None

    items = []
    for trace in xpad_idl_dict['ptr']:
        item = XPadDataItem()

        item.name = trace['NAME'][0]
        item.source = trace['SOURCE'][0]
        item.label = trace['DINFO'][0]['LABEL'][0]
        item.units = trace['DINFO'][0]['UNITS'][0]

        numdims = trace['SIZE'][0][0]

        item.data = rollaxis(trace['DATA'][0], numdims - 1)

        if numdims > 0:
            # f(t), f(t,x) or f(t,x,y)
            dim = XPadDataDim()

            # If we have a time domain, use data from TINFO
            if trace['TINFO'][0]['DOMAINS'][0] > 0:
                dim.data = linspace(trace['TINFO'][0]['START'][0],
                                    trace['TINFO'][0]['FINISH'][0],
                                    trace['TINFO'][0]['LENGTH'])
            else:
                dim.data = trace['TIME'][0]

            dim.name = trace['TINFO'][0]['LABEL'][0]
            dim.label = dim.name
            dim.units = trace['TINFO'][0]['UNITS'][0]

            item.dim.append(dim)

            item.order = len(item.dim) - 1
            item.time = item.dim[item.order].data

        if numdims > 2:
            # f(t,x,y)
            dim = XPadDataDim()
            dim.data = trace['Y'][0]
            dim.name = trace['YINFO'][0]['LABEL'][0]
            dim.label = dim.name
            dim.units = trace['YINFO'][0]['UNITS'][0]

            item.dim.append(dim)

        if numdims > 1:
            # f(t,x) or f(t,x,y)
            dim = XPadDataDim()
            dim.data = trace['X'][0]
            dim.name = trace['XINFO'][0]['LABEL'][0]
            dim.label = dim.name
            dim.units = trace['XINFO'][0]['UNITS'][0]

            item.dim.append(dim)

        item.desc = trace['TYPE'][0]

        items.append(item)

    return items


def check_padsav(xpad_idl_dict):
    """
        Verifies padsav data and its containing traces are valid.

        Parameters
        ----------
        xpad_idl_dict : AttrDict or dict
            An XPad *.padsav imported using scipy.io.readsav


        Returns
        -------
        int : 0 for invalid, 1 for valid
    """
    if not isinstance(xpad_idl_dict, dict):
        return 0

    if 'ptr' not in xpad_idl_dict:
        return 0

    # [*] 'ptr not found in xpad_idl_dict'
    if not xpad_idl_dict['ptr'].size:
        return 0

    for trace in xpad_idl_dict['ptr']:
        # [*] failed on trace
        if not check_trace(trace):
            return 0

    return 1


def check_trace(xpad_idl_trace):
    """
        Verifies a trace is well-formed and valid

        Parameters
        ----------
        xpad_idl_trace : numpy.core.records.recarray
            A trace from imported XPad *.padsav data


        Returns
        -------
        int : 0 for invalid, 1 for valid
    """
    from numpy import prod
    from numpy.core.records import recarray

    # [*] 'xpad_idl_trace is not an instance of numpy.core.records.recarray'
    if not isinstance(xpad_idl_trace, recarray):
        return 0

    # [*] 'UTYPE != \'DBstructure\''
    if 'UTYPE' not in xpad_idl_trace.dtype.names or not xpad_idl_trace['UTYPE'][0] == 'DBstructure':
        return 0

    for field in ['TYPE', 'NAME', 'DATA', 'DINFO', 'SOURCE', 'PROCESS', 'SIZE', 'TINFO']:
        # [*] 'Required field missing: '+field
        if field not in xpad_idl_trace.dtype.names:
            return 0

    # [*] TINFO.UTYPE != \'TINFO\'
    if 'UTYPE' not in xpad_idl_trace['TINFO'][0].dtype.names or not xpad_idl_trace['TINFO'][0]['UTYPE'][0] == 'TINFO':
        return 0

    for field in ['DOMAINS', 'START', 'FINISH', 'STEP', 'SAMPLES', 'LENGTH', 'UNITS', 'LABEL']:
        # [*] 'Required TINFO field missing: '+field
        if field not in xpad_idl_trace['TINFO'][0].dtype.names:
            return 0

    types = ['f(t)', 'f(t,x)', 'f(t,x,y)']
    # [*] 'TYPE is not one of '+str(types)
    if not xpad_idl_trace['TYPE'][0] in types:
        return 0

    sizes = xpad_idl_trace['SIZE'][0]
    numdims = sizes[0]
    # [*] 'size(DATA) != prod(dim_sizes)'
    if not xpad_idl_trace['DATA'][0].size == prod(sizes[1:
    ]): return 0

    # [*] 'Incorrect numdims for shape of DATA'
    if not numdims == len(xpad_idl_trace['DATA'][0].shape):
        return 0

    # [*] 'Incorrect numdims in SIZE for TYPE'
    if not numdims == types.index(xpad_idl_trace['TYPE'][0]) + 1:
        return 0

    if numdims > 0:
        if xpad_idl_trace['TINFO'][0]['DOMAINS'][0] < 1:
            if 'TIME' in xpad_idl_trace.dtype.names:
                # [*] 'size(TIME) != SIZE[1]':
                if not xpad_idl_trace['TIME'][0].size == sizes[1]:
                    return 0

                # Should this always be true?
                # [*] 'size(TIME) != TINFO.LENGTH':
                #if not xpad_idl_trace['TIME'][0].size == xpad_idl_trace['TINFO'][0]['LENGTH'][0]: return 0
            else:
                # Can't determine TIME dimension
                # [*] 'TINFO contains no domains and TIME field does not exist'
                return 0
        else:
            # [*] 'TINFO.START > TINFO.FINISH'
            if not xpad_idl_trace['TINFO'][0]['START'][0] <= xpad_idl_trace['TINFO'][0]['FINISH'][0]:
                return 0

            # [*] 'TINFO.LENGTH is < 1'
            if not xpad_idl_trace['TINFO'][0]['LENGTH'][0] >= 1:
                return 0

    if numdims > 1:
        for field in ["X", "XINFO"]:
            # [*] 'Required field missing: '+field
            if field not in xpad_idl_trace.dtype.names:
                return 0

        # [*] 'size(X) != SIZE[2]':
        if not xpad_idl_trace['X'][0].size == sizes[2]:
            return 0

    if numdims > 2:
        for field in ["Y", "YINFO"]:
            # [*] 'Required field missing: '+field
            if field not in xpad_idl_trace.dtype.names:
                return 0

        # [*] 'size(Y) != SIZE[3]':
        if not xpad_idl_trace['Y'][0].size == sizes[3]:
            return 0

    # Add more checks if you so desire
    # ...
    return 1



"""
Author: Ben Dudson, Department of Physics, University of York
        benjamin.dudson@york.ac.uk
        Edward Blair, Peter Hill, John Wilson

This file is part of PyXPad.

PyXPad is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyXPad is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
"""

from PySide.QtGui import (QAbstractItemView, QAction, QCursor,
                          QFileDialog, QIcon, QMainWindow, QMenu, QMessageBox,
                          QStyle, QTableWidgetItem, QTreeWidgetItem)
from PySide.QtCore import Qt, QTextCodec, QDir

from .pyxpad_main import Ui_MainWindow
from .configdialog import ConfigDialog

from collections import OrderedDict

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    from StringIO import StringIO  # Python 2
except ImportError:
    from io import StringIO

import sys
import os
import re
import string
import fnmatch  # For matching names to wildcard patterns
from keyword import iskeyword  # Test if a string is a keyword
import xdg                     # Names of XDG directories for config

from pyxpad import fourier         # FFT-based methods
from pyxpad import calculus        # Integration and differentiation methods
from pyxpad import user_functions  # Miscellaneous useful functions

# On some installations (Python 2.6 only?) need to convert text using
# QTextCodec
codec = QTextCodec.codecForName("UTF-8")
def toStr(uni):
    return str(codec.fromUnicode(uni))


class Sources:
    sources = []  # List of sources

    def __init__(self, mainwindow):
        self.main = mainwindow
        self.main.sourceDescription.stateChanged.connect(self.updateDisplay)

        self.groupIcon = QIcon()
        self.groupIcon.addPixmap(mainwindow.style().standardPixmap(QStyle.SP_DirClosedIcon),
                                 QIcon.Normal, QIcon.Off)
        self.groupIcon.addPixmap(mainwindow.style().standardPixmap(QStyle.SP_DirOpenIcon),
                                 QIcon.Normal, QIcon.On)
        self.keyIcon = QIcon()
        self.keyIcon.addPixmap(mainwindow.style().standardPixmap(QStyle.SP_FileIcon))

        self.main.treeView.itemSelectionChanged.connect(self.updateDisplay)
        self.main.tracePattern.returnPressed.connect(self.updateDisplay)
        self.main.treeView.setContextMenuPolicy(Qt.CustomContextMenu)  # Enable popup menus

        # Context menu
        self.main.treeView.customContextMenuRequested.connect(self.handlePopupMenu)

        # Context menu actions
        self.actionAdd = QAction("Add", self.main, statusTip="Add a new source")

        self.actionDelete = QAction("Delete", self.main, statusTip="Remove source from tree")
        self.actionDelete.triggered.connect(self.deleteSource)

        self.actionConfig = QAction("Configure", self.main, statusTip="Configure source")
        self.actionConfig.triggered.connect(self.configureSource)

    def saveState(self, f):
        pickle.dump(self.sources, f)

    def loadState(self, f):
        try:
            sources = pickle.load(f)
        except EOFError:
            sources = []
        for s in sources:
            self.addSource(s)
        self.updateDisplay()

    def handlePopupMenu(self):
        """
        Called when user right-clicks on the sources tree
        """
        menu = QMenu()
        menu.addAction(self.actionAdd)

        selected = self.main.treeView.selectedItems()
        if len(selected) != 0:
            if 'config' in selected[0].source.__dict__:
                menu.addAction(self.actionConfig)
            menu.addAction(self.actionDelete)

        menu.exec_(QCursor.pos())

    def addNetCDF(self):
        """
        Add a NetCDF file as a data source
        """
        try:
            from pyxpad.datafile import NetCDFDataSource
        except ImportError:
            self.main.write("Sorry, no NetCDF support")
            return
        try:
            # Get the file name
            tr = self.main.tr
            fname, _ = QFileDialog.getOpenFileName(self.main, tr('Open file'), '.',
                                                   filter=tr("NetCDF files (*.nc *.cdl)"))
            fname = toStr(fname)
            if (fname is None) or (fname == ""):
                return  # Cancelled

            s = NetCDFDataSource(fname)

            self.addSource(s)
            self.updateDisplay()
        except:
            self.main.write("Error creating NetCDFDataSource")
            self.main.write(str(sys.exc_info()))

    def addXPADtree(self):
        try:
            from pyxpad.xpadsource import XPadSource
        except ImportError:
            self.main.write("Sorry, no XPAD tree support")
            self.main.write(str(sys.exc_info()))
            return

        try:
            # Select the directory
            tr = self.main.tr
            dname = QFileDialog.getExistingDirectory(self.main, tr('Open XPAD directory'),
                                                     QDir.currentPath())
            dname = toStr(dname)
            if (dname == "") or (dname is None):
                return
            # Create data source
            s = XPadSource(dname)

            # Add data source and update
            self.addSource(s)
            self.updateDisplay()
        except:
            self.main.write("Error creating XPadSource")
            self.main.write(str(sys.exc_info()))
            raise

    def addBOUT(self):
        """
        Add a BOUT++ directory source
        """
        try:
            from pyxpad.boutsource import BoutDataSource

            # Select the directory
            tr = self.main.tr
            dname = QFileDialog.getExistingDirectory(self.main, tr('Open BOUT++ directory'),
                                                     QDir.currentPath())
            dname = toStr(dname)
            if (dname == "") or (dname is None):
                return
            # Create data source
            s = BoutDataSource(dname)

            # Add data source and update
            self.addSource(s)
            self.updateDisplay()
        except:
            self.main.write("Sorry, no BOUT++ support")
            raise
            return

    def addSource(self, source):
        self.sources.append(source)
        it = QTreeWidgetItem(self.main.treeView, [source.label])
        it.setIcon(0, self.groupIcon)
        it.source = source
        self.main.treeView.addTopLevelItem(it)

        def buildtree(parent, it):
            # Check for children
            try:
                for child in parent.children:
                    itchild = QTreeWidgetItem(it, [child.label])
                    itchild.source = child
                    buildtree(child, itchild)  # Add child's children
                    it.addChild(itchild)
            except AttributeError:
                # Probably no children
                return

        buildtree(source, it)

    def deleteSource(self):
        tree = self.main.treeView
        selected = tree.selectedItems()
        if len(selected) == 0:
            return
        source = selected[0].source
        tree.takeTopLevelItem(tree.indexOfTopLevelItem(selected[0]))  # Remove from tree
        # Remove from list of sources
        i = self.sources.index(source)
        del self.sources[i]
        # Update the display
        self.updateDisplay()

    def configureSource(self):
        """
        Configure a data source, changing the source's
        'config' dictionary.
        """
        selected = self.main.treeView.selectedItems()
        if len(selected) == 0:
            return
        source = selected[0].source
        c = ConfigDialog(source.config, self.main)
        c.exec_()

    def updateDisplay(self):
        table = self.main.sourceTable
        # Find which source is selected, and update table view
        selected = self.main.treeView.selectedItems()
        if len(selected) == 0:
            table.clearContents()
            table.setRowCount(0)
            return
        s = selected[0].source

        # Check if any items selected
        selecteditems = table.selectedItems()
        selectedvars = []
        nextra = 0
        for item in selecteditems:
            if 'source' in item.__dict__:
                name = toStr(item.text())
                selectedvars.append((name, item.source))
                if item.source != s:
                    nextra += 1

        table.clearContents()  # Clear the table and selections

        pattern = toStr(self.main.tracePattern.text())
        if pattern == "":
            varNames = s.varNames
        else:
            # Filter the variable names
            varNames = [name for name in s.varNames
                        if fnmatch.fnmatch(name.lower(), pattern.lower())]

        varNames.sort(key=str.lower)

        if self.main.sourceDescription.isChecked():
            # Provide description for each variable (if available)

            table.setColumnCount(2)
            table.setRowCount(len(varNames) + nextra)
            table.setSelectionBehavior(QAbstractItemView.SelectRows)

            def addVar(name, source, selected=False):
                var = source.variables[name]
                item = QTableWidgetItem(name)
                item.source = source
                table.setItem(addVar.ind, 0, item)
                item.setSelected(selected)
                comment = var.desc
                if comment == "":
                    comment = var.label
                    if var.units != "":
                        comment += " ("+var.units+") "

                if var.dim:
                    try:
                        comment += " [" + ", ".join([str(v) for v in var.dim]) + "]"
                    except TypeError:
                        if str(var.dim):
                            comment += " [" + str(var.dim) + "]"
                item = QTableWidgetItem(comment)
                table.setItem(addVar.ind, 1, item)
                item.setSelected(selected)
                addVar.ind += 1
        else:
            # Just a list of variable names. Can use multiple columns
            maxrows = 20
            n = len(varNames) + nextra
            ncols = int(n / maxrows) + 1
            table.setColumnCount(ncols)
            table.setRowCount(min([n, maxrows]))
            table.setSelectionBehavior(QAbstractItemView.SelectItems)

            def addVar(name, source, selected=False):
                row = addVar.ind % maxrows
                col = int(addVar.ind / maxrows)
                item = QTableWidgetItem(name)
                item.source = source
                table.setItem(row, col, item)
                item.setSelected(selected)
                addVar.ind += 1

        addVar.ind = 0
        for name, source in selectedvars:
            addVar(name, source, True)
        sel = [name for name, source in selectedvars if source == s]
        for name in varNames:
            if name not in sel:
                addVar(name, s)

    def read(self):
        """
        Read the selected data and return as a list of data items

        Input
        -----
            None

        Returns
        ------
            [ XPadDataItem ]  or equivalent

        Modifies
        --------
            None
        """

        # Get list of shots
        shotlist = toStr(self.main.shotInput.text()).split(',')

        table = self.main.sourceTable
        tableitems = table.selectedItems()
        data = []
        for item in tableitems:
            if 'source' in item.__dict__:
                name = toStr(item.text())

                for shot in shotlist:
                    s = "Reading " + name + " from " + item.source.label
                    if shot != "":
                        s += " shot = " + shot
                    self.main.write(s)
                    # Run in a sandbox to catch exceptions and display output
                    self.main.runSandboxed(lambda: data.append(item.source.read(name, shot)))

            else:
                print("Ignoring "+item.text())
        return data


class PyXPad(QMainWindow, Ui_MainWindow):
    """

    Attributes

      data    Dictionary of variables containing user data
    """
    def __init__(self, parent=None, loadfile=None, ignoreconfig=False):
        super().__init__(parent)
        self.setupUi(self)

        self.sources = Sources(self)  # Handles data sources
        self.data = OrderedDict()  # User data

        # File menu
        self.actionNetCDF_file.triggered.connect(self.sources.addNetCDF)
        self.actionXPAD_tree.triggered.connect(self.sources.addXPADtree)
        self.actionBOUT_data.triggered.connect(self.sources.addBOUT)

        self.actionExit.triggered.connect(self.close)

        self.actionLoadState.triggered.connect(self.loadState)
        self.actionSaveState.triggered.connect(self.saveState)

        # Graphics menu
        self.actionPlot.triggered.connect(self.handlePlot)
        self.actionOPlot.triggered.connect(self.handleOPlot)
        self.actionMPlot.triggered.connect(self.handleMPlot)
        self.actionXYPlot.triggered.connect(self.handleXYPlot)
        self.actionZPlot.triggered.connect(self.handleZPlot)
        self.actionContour.triggered.connect(self.handleContour)
        self.actionContour_filled.triggered.connect(self.handleContourf)
        self.actionClearFig.triggered.connect(self.handleClearFig)

        # Command menu
        self.actionDeleteTrace.triggered.connect(self.handleDeleteTrace)
        self.actionChop.triggered.connect(self.handleChop)
        self.actionIntegrate.triggered.connect(self.handleIntegrate)
        self.actionDf_dt.triggered.connect(self.handleDifferentiate)
        self.actionAdd.triggered.connect(self.handleAdd)
        self.actionMultiply.triggered.connect(self.handleMultiply)
        self.actionSubtract.triggered.connect(self.handleSubtract)
        self.actionDivide.triggered.connect(self.handleDivide)
        self.actionFFTP.triggered.connect(self.handleFFTP)
        self.actionRunFFT.triggered.connect(self.handleRunFFT)
        self.actionReciprocal.triggered.connect(self.handleReciprocal)
        self.actionExponential.triggered.connect(self.handleExponential)
        self.actionAbsolute.triggered.connect(self.handleAbsolute)
        self.actionArctan.triggered.connect(self.handleArctan)
        self.actionNlog.triggered.connect(self.handleNlog)
        self.actionNorm.triggered.connect(self.handleNorm)
        self.actionInvert.triggered.connect(self.handleInvert)
        self.actionAddCon.triggered.connect(self.handleAddCon)
        self.actionSubCon.triggered.connect(self.handleSubCon)
        self.actionMulCon.triggered.connect(self.handleMulCon)
        self.actionDivCon.triggered.connect(self.handleDivCon)
        self.actionPowCon.triggered.connect(self.handlePowCon)
        self.actionChangeName.triggered.connect(self.handleChangeName)
        self.actionChangeUnits.triggered.connect(self.handleChangeUnit)
        self.actionClip.triggered.connect(self.handleClip)
        self.actionStats.triggered.connect(self.handleStats)
        self.actionTimeOff.triggered.connect(self.handleTimeOff)

        # Help menu
        self.actionAbout.triggered.connect(self.handleAbout)

        # Sources tab
        self.readDataButton.clicked.connect(self.readData)
        self.shotInput.returnPressed.connect(self.readData)
        self.lastShotButton.clicked.connect(self.lastShot)

        self.commandInput.commandEntered.connect(self.commandEntered)
        self.commandButton.clicked.connect(self.commandEntered)

        # Data tab
        self.dataTable.cellChanged.connect(self.dataTableChanged)
        self.dataTable.customContextMenuRequested.connect(self.handlePopupMenu)

        try:
            from pyxpad.matplotlib_widget import MatplotlibWidget
            self.DataPlot = MatplotlibWidget(self.plotTab)
        except:
            raise

        if not ignoreconfig and loadfile is None:
            # Other configuration can be saved in here
            self.config_dir = os.path.join(xdg.XDG_CONFIG_HOME, 'pyxpad')
            if os.path.exists(xdg.XDG_CACHE_HOME):
                os.makedirs(self.config_dir, exist_ok=True)
                defaultfile = os.path.join(self.config_dir, "saved_state.pyx")
                if os.path.exists(defaultfile):
                    loadfile = defaultfile
        else:
            self.config_dir = os.getcwd()

        # Load state
        if loadfile is not None:
            self.config_dir = os.path.dirname(os.path.abspath(loadfile))
            self.loadState(loadfile)

    def saveState(self, filename=None):
        """
        Saves program state to given file. If no file is specified,
        then a dialog is created to ask the user for one.
        """
        if filename is None:
            tr = self.tr
            defaultfile = os.path.join(self.config_dir, "saved_state.pyx")
            filename, _ = QFileDialog.getSaveFileName(self, dir=defaultfile,
                                                      filter=tr("PyXPad save file (*.pyx)"))
            filename = toStr(filename)
        if (filename is None) or (filename == ""):
            return
        try:
            with open(filename, 'wb') as f:
                self.sources.saveState(f)
                pickle.dump(self.data, f)
            self.write("** Saved state to file '"+filename+"'")
        except:
            e = sys.exc_info()
            self.write("Could not save state to file '"+filename+"'")
            self.write("\t ->" + str(e[1]))

    def loadState(self, filename=None):
        """
        Loads program state from the given filename.
        If no filename is specified, then a dialog is created
        to ask the user for a file name.
        """
        if filename is None:
            tr = self.tr
            filename, _ = QFileDialog.getOpenFileName(self, tr('Open file'), '.',
                                                      filter=tr("PyXPad save file (*.pyx)"))
            filename = toStr(filename)
        if (filename is None) or (filename == ""):
            return  # Cancelled
        if not os.path.exists(filename):
            self.write("Could not find " + filename)
            return
        try:
            with open(filename, 'rb') as f:
                self.sources.loadState(f)
                self.data = pickle.load(f)
        except EOFError:
            self.data = OrderedDict()
        except:
            e = sys.exc_info()
            self.write("Could not load state from file '"+filename+"'")
            self.write("\t ->" + str(e[1]))
            raise
        else:
            # If no exception raised, then update tables, lists
            self.data = OrderedDict(self.data)
            self.updateDataTable()
            self.write("** Loaded state from file '"+filename+"'")

    def closeEvent(self, event):
        """
        Called when the main window is closed
        """
        reply = QMessageBox.question(self, 'Message',
                                           "Are you sure to quit?", QMessageBox.Yes |
                                           QMessageBox.No, QMessageBox.Yes)

        if reply == QMessageBox.Yes:
            event.accept()
        else:
            event.ignore()

    def write(self, text):
        """
        Write some log text to output text widget
        """
        self.textOutput.append(text)

    def makeUnique(self, name):
        """
        Modifies a given string into a valid Python variable name
        which is not already in self.data

        Input
        -----
            name  ::  string

            self.data   (class member)

        Returns
        ------
            string containing modified name

        Modifies
        --------
            None
        """
        # First make sure the name is a valid variable name
        name = re.sub('\W|^(?=\d)', '_', name)  # Replace invalid characters with '_'

        if iskeyword(name):  # Check if name is a keyword
            name += '2'

        if name in self.data:
            # Name is already in the list. Add a number to the end to make it unique
            i = 1
            while name + "_"+str(i) in self.data:
                i += 1
            return name + "_"+str(i)
        return name

    def uniqueName(self):
        """
        Generates a unique variable name, which
        is not already in self.data
        """

        def findName(name, length):
            """
            Finds a name
            """
            for c in string.ascii_lowercase:
                if length <= 1:
                    if name+c not in self.data:
                        return name+c
                else:
                    r = findName(name+c, length-1)
                    if r is not None:
                        return r
            return None

        length = 1
        while True:
            n = findName("", length)
            if n is not None:
                return n
            length += 1

    def readData(self):
        """
        User pressed the "Read" button to read selected
        data items from given shots.

        Input
        -----
            None

        Returns
        ------
            None

        Modifies
        --------
            self.data


        Calls Sources to read the new data

        Inserts data into self.data, ensuring that the name
        of each data item is unique and a valid Python name

        Updates the data table based on self.data
        """
        # Switch to data tab
        self.tabWidget.setCurrentWidget(self.dataTab)
        # Get the data from the source as a list
        newdata = self.sources.read()
        if (newdata is None) or (newdata == []):
            return  # No data read

        # Add to the data dictionary
        for item in newdata:
            try:
                # Need to make the name unique
                name = self.makeUnique(item.name)
                self.data[name] = item
            except:
                self.write("Error adding item '"+str(item)+"'")

        self.updateDataTable()

    def lastShot(self):
        """
        Get the latest shot number
        """

        # Find an XPadSource in the list of sources
        from pyxpad.xpadsource import XPadSource
        source = None
        for _source in self.sources.sources:
            if isinstance(_source, XPadSource):
                source = _source
        if source is None:
            # None available
            return

        # Now ask for the lastshot
        data_item = source.read("lastshot", "")
        last_shot_number = data_item.data.children[0].lastshot

        # Append the shot number to any existing shot numbers
        current_text = self.sources.main.shotInput.text()
        if current_text == '':
            new_text = str(last_shot_number)
        else:
            new_text = ', '.join([current_text, str(last_shot_number)])
        self.sources.main.shotInput.setText(new_text)

        return last_shot_number

    def updateDataTable(self):
        """
        Updates the table of data based on self.data dictionary
        """
        n = len(self.data)
        table = self.dataTable
        table.setSortingEnabled(False)  # Stops the table rearranging itself
        self.dataTable.cellChanged.disconnect(self.dataTableChanged)  # Don't call the dataTableChanged function
        table.setRowCount(n)
        for row, name in enumerate(self.data):
            item = self.data[name]
            it = QTableWidgetItem(name)
            it.oldname = name  # Save this for when it's changed
            table.setItem(row, 0, it)

            # Assume it's an XPadDataItem
            try:
                it = QTableWidgetItem(item.source)
                it.setFlags(it.flags() ^ Qt.ItemIsEditable)  # Make source read only
                table.setItem(row, 1, it)
            except:
                table.setItem(row, 1, QTableWidgetItem(""))

            try:
                it = QTableWidgetItem(item.name)
                it.setFlags(it.flags() ^ Qt.ItemIsEditable)  # Make trace read only
                table.setItem(row, 2, it)
            except:
                table.setItem(row, 2, QTableWidgetItem(""))

            try:
                try:
                    comment = item.comment
                except AttributeError:
                    comment = item.desc
                    if comment == "":
                        comment = item.label
                        if item.units != "":
                            comment += " ("+item.units+") "

                    if item.dim != []:
                        comment += " [" + item.dim[0].name
                        for d in item.dim[1:]:
                            comment += ", " + d.name
                        comment += "] "
                    else:
                        comment += " = " + str(item.data)

                table.setItem(row, 3, QTableWidgetItem(comment))
            except:
                table.setItem(row, 3, QTableWidgetItem(str(item)))

        table.setSortingEnabled(True)  # Re-enable sorting
        self.dataTable.cellChanged.connect(self.dataTableChanged)

    def dataTableChanged(self, row, col):
        """
        Called when the user changes the value of a cell
        in the data table. This can either be to change
        the name of a variable, or the comment.
        """
        if col == 0:
            # The name of the variable
            it = self.dataTable.item(row, col)
            name = toStr(it.text())
            oldname = it.oldname
            if name == oldname:
                return  # Not really changed

            # Need to make sure new name is unique and valid
            name = self.makeUnique(name)
            it.setText(name)
            it.oldname = name
            self.data[name] = self.data[oldname]
            del self.data[oldname]

        if col == 3:
            # Changing the comment
            comment = self.dataTable.item(row, col).text()
            name = self.dataTable.item(row, 0).text()
            self.data[name].comment = comment

        print(row, col)

    def commandEntered(self, text=None):
        """
        Called when a command is entered on the Data tab.
        Gets the command string from the text box, and calls
        the runCommand to run the command.
        """
        # If there's no text, then the "Run" button was probably
        # pressed
        if text is None:
            text = self.commandInput.text()

        # Get the command text and clear the text box
        cmd = toStr(text)
        self.commandInput.clear()
        self.runCommand(cmd)

    def selectedDataNames(self):
        """
        Retuns a list of the names of variables selected
        in the Data table.
        """
        # Find which items are selected in the data table
        items = self.dataTable.selectedItems()  # All selected items
        if len(items) == 0:
            return []
        names = []  # List of selected data names
        for it in items:
            try:
                names.append(it.oldname)
            except AttributeError:
                pass  # Ignore if doesn't have a name
        return names

    def handlePopupMenu(self):
        """
        Called when user right-clicks on a trace
        """
        menu = QMenu()

        selected = self.selectedDataNames()
        if len(selected) != 0:
            menu.addAction(self.actionDeleteTrace)

        menu.exec_(QCursor.pos())

    ##################### Plot menu actions #####################

    def handlePlot(self):
        # Find which items are selected
        names = self.selectedDataNames()
        if len(names) == 0:
            return

        # Sort by trace name
        def trace(name):
            try:
                return self.data[name].name
            except:
                return ""
        namelist = [(name, trace(name)) for name in names]
        values = set(map(lambda x: x[1], namelist))
        groups = [[y[0] for y in namelist if y[1] == x] for x in values]

        def plotStr(items):
            return "[" + ", ".join(items) + "]"

        # Create a command to execute
        cmd = "plot(" + ", ".join([plotStr(group) for group in groups]) + ")"

        # Run the command within sandboxed environment
        # Also shows user the command which can be entered
        self.runCommand(cmd)
        self.tabWidget.setCurrentWidget(self.plotTab)

    def handleOPlot(self):
        """
        Creates an overlap plot of selected traces
        """

        names = self.selectedDataNames()

        if len(names) == 0:
            return

        def trace(name):
            try:
                return self.data[name].name
            except:
                return ""
        namelist = [(name, trace(name)) for name in names]
        values = set(map(lambda x: x[1], namelist))
        groups = [[y[0] for y in namelist if y[1] == x] for x in values]

        def plotStr(items):
            return "[" + ", ".join(items) + "]"

        cmd = "oplot(" + ", ".join([plotStr(group) for group in groups]) + ")"
        self.runCommand(cmd)
        self.tabWidget.setCurrentWidget(self.plotTab)

    def handleMPlot(self):
        """
        Creates multiple subplots of multiple traces input by the user
        """

        names = self.selectedDataNames()

        if len(names) == 0:
            return

        def trace(name):
            try:
                return self.data[name].name
            except:
                return ""

        namelist = [(name, trace(name)) for name in names]
        values = set(map(lambda x: x[1], namelist))
        groups = [[y[0] for y in namelist if y[1] == x] for x in values]

        def plotStr(items):
            return "[" + ", ".join(items) + "]"

        cmd = "mplot(" + ", ".join([plotStr(group) for group in groups]) + ")"
        self.runCommand(cmd)
        self.tabWidget.setCurrentWidget(self.plotTab)

    def handleXYPlot(self):
        names = self.selectedDataNames()
        if len(names) != 2:
            self.write("** Two data items must be selected for X-Y plotting")
            return
        # Run the command in sandbox
        self.runCommand("plotxy( "+names[0]+", "+names[1]+")")
        self.tabWidget.setCurrentWidget(self.plotTab)

    def handleZPlot(self):
        """
        Creates an zoomed plot of selected traces
        """

        names = self.selectedDataNames()

        if len(names) == 0:
            return

        def trace(name):
            try:
                return self.data[name].name
            except:
                return ""
        namelist = [(name, trace(name)) for name in names]
        values = set(map(lambda x: x[1], namelist))
        groups = [[y[0] for y in namelist if y[1] == x] for x in values]

        def plotStr(items):
            return "[" + ", ".join(items) + "]"

        cmd = "zplot(" + ", ".join([plotStr(group) for group in groups]) + ")"
        self.runCommand(cmd)
        self.tabWidget.setCurrentWidget(self.plotTab)

    def handleClearFig(self):
        self.runCommand("clearFig()")

    def handleContour(self):
        """
        Make a contour plot of a 2D trace
        """
        names = self.selectedDataNames()
        if len(names) != 1:
            self.write("** One data item must be selected for contour")
            return

        self.runCommand("contour("+names[0]+")")
        self.tabWidget.setCurrentWidget(self.plotTab)

    def handleContourf(self):
        """
        Make a contour plot of a 2D trace
        """
        names = self.selectedDataNames()
        if len(names) != 1:
            self.write("** One data item must be selected for contourf")
            return

        self.runCommand("contourf("+names[0]+")")
        self.tabWidget.setCurrentWidget(self.plotTab)

    def handleCommandAction(self, command):
        names = self.selectedDataNames()
        if len(names) != 1:
            self.write("** One data item must be selected for "+command)
            return
        # Run the command in sandbox
        self.runCommand(command+"( "+names[0]+" )")

    ######## Command menu handlers

    def handleDeleteTrace(self):
        """
        Delete selected traces
        """
        # Get list of selected variables
        names = self.selectedDataNames()

        if len(names) == 0:
            return

        for name in names:
            self.runCommand("del({})".format(name))

    def handleChop(self):
        """
        Chops a signal, keeping only specified time range
        """
        # Get list of selected variables
        names = self.selectedDataNames()

        if len(names) == 0:
            return

        try:
            # Get current time range from first variable
            var = self.data[names[0]]
            if var.time is not None:
                tmin = var.time[0]
                tmax = var.time[-1]
            else:
                tmin = var.dim[0].data[0]
                tmax = var.dim[0].data[-1]
        except:
            return

        # Use a dialog box to get time range
        config = OrderedDict(("min", float(tmin)), ("max", float(tmax)))

        c = ConfigDialog(config, self)
        c.exec_()

        for n in names:
            self.runCommand(self.makeUnique(n+"_chop") + " = " +
                            "chop( "+n+", %e, %e )" % (config["min"], config["max"]))

    def handleIntegrate(self):
        """
        Integrates one or more traces
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "intg( "+n+" )")

    def handleDifferentiate(self):
        """
        Differentiates one or more traces
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "diff(" + n + ")")

    def handleAdd(self):
        """
        Adds all selected traces together
        """
        names = self.selectedDataNames()
        if len(names) < 2:
            self.write("** At least two data items must be selected to add together")
            return

        self.runCommand(self.uniqueName() + " = " + "+".join(names))

    def handleMultiply(self):
        """
        Adds all selected traces together
        """
        names = self.selectedDataNames()
        if len(names) < 2:
            self.write("** At least two data items must be selected to multiply together")
            return

        self.runCommand(self.uniqueName() + " = " + "*".join(names))

    def handleSubtract(self):
        """
        Subtracts one trace from another

        """
        names = self.selectedDataNames()
        if len(names) != 2:
            self.write("** Two data items must be selected to subtract one from another")
            return
        self.runCommand(self.uniqueName()+" = " + names[0] + " - " + names[1])

    def handleDivide(self):
        """
        Divide one trace by another
        """
        names = self.selectedDataNames()
        if len(names) != 2:
            self.write("** Two data items must be selected to subtract one from another")
            return
        self.runCommand(self.uniqueName()+" = " + names[0] + " / " + names[1])

    def handleFFTP(self):
        """
        Perform FFT, returning amplitude and phase

        """

        for n in self.selectedDataNames():
            # Create a unique name for amplitude and phase
            ampname = self.makeUnique(n+"_amp")
            phasename = self.makeUnique(n+"_phase")
            self.runCommand(ampname+","+phasename + " = " + "fftp( "+n+" )")

    def handleRunFFT(self):
        """
        Perform Running FFT
        """

        names = self.selectedDataNames()

        if len(names) == 0:
            return

        stride = 0.001
        width = 0.001

        # Use a dialog box to get width and stride
        config = (("stride", float(width)), ("width", float(stride)))
        config = OrderedDict(config)

        c = ConfigDialog(config, self)
        c.exec_()

        for name in names:
            # Create a unique name
            new_name = self.makeUnique(name + "_runfft")
            self.runCommand(new_name + " = " +
                            "runfft({}, stride={}, width={})".format(name, config["stride"], config["width"]))

    def handleReciprocal(self):
        """
        Returns the reciprocal of one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "recip( "+n+" )")

    def handleExponential(self):
        """
        Returns the reciprocal of one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "exp( "+n+" )")

    def handleAbsolute(self):
        """
        Returns the absolute value of one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "abs( "+n+" )")

    def handleArctan(self):
        """
        Returns the arctan of one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "atan( "+n+" )")

    def handleNlog(self):
        """
        Returns the natural log of one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "ln( "+n+" )")

    def handleNorm(self):
        """
        Normalises and returns one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "norm( "+n+" )")

    def handleInvert(self):
        """
        Returns the inversion of one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "inv( "+n+" )")

    def handleAddCon(self):
        """
        Adds a constant to and returns one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "addcons( "+n+" )")

    def handleSubCon(self):
        """
        Subtracts a constant from and returns one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "subcons( "+n+" )")

    def handleMulCon(self):
        """
        Multiplies by a constant and returns one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "mulcons( "+n+" )")

    def handleDivCon(self):
        """
        Divides by a constant and returns one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "divcons( "+n+" )")

    def handlePowCon(self):
        """
        Raises to the power of a constant and returns one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "powcons( "+n+" )")

    def handleChangeName(self):
        """
        Changes the name of one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            name = user_functions.inputname()
            self.runCommand(self.makeUnique(name) + " = " + "renamed( "+n+" )")

    def handleChangeUnit(self):
        """
        Changes the units of one or more trace(s)
        """
        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "newunits( "+n+" )")

    def handleClip(self):
        """
        Clips a signal, keeping only a specified value ranged
        """

        names = self.selectedDataNames()

        if len(names) == 0:
            return

        try:
            # Get current data range from first variable
            var = self.data[names[0]]
            valmin = var.data[0]
            valmax = var.data[0]
            for point in var.data:
                if point < valmin:
                    valmin = point
                if point > valmax:
                    valmax = point
        except:
            return

        # Use a dialog box to get value range
        config = OrderedDict(("min", float(valmin)), ("max", float(valmax)))

        c = ConfigDialog(config, self, pstvOnly=False)
        c.exec_()

        for n in names:
            self.runCommand(self.makeUnique(n+"_clip") + " = " +
                            "clip("+n+", %e, %e )" % (config["min"], config["max"]))

    def handleStats(self):
        """
        Returns the statistics (mean, standard deviation and range) of one or more traces
        """

        names = self.selectedDataNames()

        if len(names) == 0:
            return

        for n in names:
            self.runCommand("stats("+n+")")

    def handleTimeOff(self):
        """
        Adds a time offset to one or more traces
        """

        names = self.selectedDataNames()

        for n in names:
            self.runCommand(self.uniqueName() + " = " + "timoff("+n+")")

    ########## Help menu handlers ##########

    def handleAbout(self):
        """
        Displays the About dialog
        """

        about_box = QMessageBox()
        about_box.setText(__doc__)
        about_box.exec_()

    ##########

    def runSandboxed(self, func, args=()):
        # To capture print statements stdout is temporarily directed to a StringIO buffer
        buffer = StringIO()
        oldstdout = sys.stdout
        sys.stdout = buffer
        val = None
        try:
            val = func(*args)
        except:
            e = sys.exc_info()
            self.write("Error: " + str(e[0]))
            self.write("Reason: " + str(e[1]))
        sys.stdout = oldstdout
        output = buffer.getvalue()
        if len(output) > 0:
            self.write(output)
        return val

    def _runExec(self, cmd, glob, loc):
        """
        This is a wrapper around exec
        Needed because exec isn't allowed in a lambda or nested function
        and can't be passed as a function pointer.
        """
        exec(cmd, glob, loc)

    def runCommand(self, cmd):
        # Output the command
        self.write(">>> " + cmd)

        glob = globals()
        glob['plot']     = self.DataPlot.plot
        glob['oplot']    = self.DataPlot.oplot
        glob['mplot']    = self.DataPlot.mplot
        glob['plotxy']   = self.DataPlot.plotxy
        glob['zplot']    = self.DataPlot.zplot
        glob['contour']  = self.DataPlot.contour
        glob['contourf'] = self.DataPlot.contourf
        glob['clearFig'] = self.DataPlot.clearFig

        glob['intg']     = calculus.integrate
        glob['diff']     = calculus.differentiate
        glob['fftp']     = fourier.fftp
        glob['runfft']   = fourier.runfft
        glob['chop']     = user_functions.chop
        glob['recip']    = user_functions.reciprocal
        glob['exp']      = user_functions.exponential
        glob['abs']      = user_functions.absolute
        glob['atan']     = user_functions.arctan
        glob['ln']       = user_functions.nlog
        glob['norm']     = user_functions.normalise
        glob['inv']      = user_functions.invert
        glob['addcons']  = user_functions.addcon
        glob['subcons']  = user_functions.subcon
        glob['mulcons']  = user_functions.mulcon
        glob['divcons']  = user_functions.divcon
        glob['powcons']  = user_functions.powcon
        glob['renamed']  = user_functions.changename
        glob['newunits'] = user_functions.changeunits
        glob['clip']     = user_functions.clip
        glob['stats']    = user_functions.statistics
        glob['timoff']   = user_functions.timeOffset

        # Evaluate the command, catching any exceptions
        # Local scope is set to self.data to allow access to user data
        self.runSandboxed(self._runExec, args=(cmd, glob, self.data))
        self.updateDataTable()



#
# Initial code taken from  http://stackoverflow.com/questions/6723527/getting-pyside-to-work-with-matplotlib
# Additional bits from https://gist.github.com/jfburkhart/2423179
# zplot zoom effect from http://matplotlib.org/examples/pylab_examples/axes_zoom_effect.html
#

import matplotlib

matplotlib.use('Qt4Agg')
matplotlib.rcParams['backend.qt4'] = 'PySide'

from matplotlib.figure import Figure
from matplotlib.transforms import Bbox, TransformedBbox, \
    blended_transform_factory
from mpl_toolkits.axes_grid1.inset_locator import BboxPatch, BboxConnector,\
    BboxConnectorPatch

from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
try:
    from matplotlib.backends.backend_qt4agg import NavigationToolbar2QTAgg as NavigationToolbar
except ImportError:
    from matplotlib.backends.backend_qt4agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.pyplot import setp

from PySide.QtGui import QVBoxLayout
from PySide.QtGui import QDialog, QGridLayout, QDialogButtonBox, QPushButton
from PySide.QtCore import Qt
from PySide.QtGui import QInputDialog

import numpy as np


class MatplotlibWidget():

    def __init__(self, parent):

        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        self.canvas.setParent(parent)
        self.mpl_toolbar = NavigationToolbar(self.canvas, parent)
        self.axes = self.figure.add_subplot(111)

        self.grid_layout = QVBoxLayout()
        self.grid_layout.addWidget(self.canvas)
        self.grid_layout.addWidget(self.mpl_toolbar)
        parent.setLayout(self.grid_layout)

        self.callback_id = None

    def _clean_axes(self):
        """
        Make sure the figure is in a nice state
        """
        # Get rid of any extra axes
        if isinstance(self.axes, list):
            for axes in self.axes:
                del(axes)
            self.axes = self.figure.add_subplot(111)
        else:
            self.axes.clear()

        self.figure.clear()
        self.axes.grid(True)
        # Reset to some hardcoded default values
        self.figure.subplots_adjust(left=0.125, right=0.9, top=0.9, bottom=0.1,
                                    wspace=0.2, hspace=0.2)
        # Remove any event callbacks
        if self.callback_id:
            try:
                self.figure.canvas.mpl_disconnect(self.callback_id)
            except TypeError:
                for callback_id in self.callback_id:
                    self.figure.canvas.mpl_disconnect(callback_id)
                self.callback_id = None

    def plot(self, *args):
        """
        Make multiple plots
        """
        nplots = len(args)
        self._clean_axes()

        for plotnum, p in enumerate(args, 1):
            # For each plot
            if plotnum == 1:
                self.axes = self.figure.add_subplot(nplots, 1, plotnum)
                ax = self.axes
            else:
                self.axes = self.figure.add_subplot(nplots, 1, plotnum, sharex=ax)
            setp(self.axes.get_xticklabels(), visible=False)
            try:
                # Assume each p is a list of data items to be overplotted
                for data in p:
                    # Plot data
                    #self.axes.plot(data.time, data.data)
                    label = data.desc
                    if label == "":
                        label = data.label
                    label += " " + data.source

                    time = data.time
                    if time is None:
                        if len(data.dim) != 1:
                            print(data.dim)
                            raise ValueError("Cannot plot '"+data.label+"' as it has too many dimensions")
                        time = data.dim[0].data

                    self.axes.plot(time, data.data, label=label)

                # Y label from last plot
                ylabel = data.desc
                if ylabel == "":
                    ylabel = data.label
                    if data.units != "":
                        ylabel += " ("+data.units+")"
                self.axes.set_ylabel(ylabel)

                if plotnum == nplots:
                    self.axes.set_xlabel(p[0].dim[p[0].order].label)
                    setp(self.axes.get_xticklabels(), visible=True)
                self.axes.legend()
            except TypeError:
                # p not iterable, so just plot item
                #self.axes.plot(p.time, p.data)
                time = p.time.data
                xlabel = p.dim[p.order].label
                if time is None:
                    if len(p.dim) != 1:
                        raise ValueError("Cannot plot '"+p.label+"' as it has too many dimensions")
                    time = p.dim[0].data
                    xlabel = p.dim[0].label

                data = p.data
                # Check the size of the array
                size = len(time)
                if size > 10000:
                    fac = int(len(time) / 10000)
                    time = time[::fac]
                    data = data[::fac]
                    print("Warning: too many samples (%d). Down-sampling to %d points" % (size, len(time)))

                self.axes.plot(time, data)

                ylabel = p.desc
                if ylabel == "":
                    ylabel = p.label
                    if p.units != "":
                        ylabel += " ("+p.units+")"
                self.axes.set_ylabel(ylabel)
                if plotnum == 0:
                    self.axes.set_xlabel(xlabel)

        self.figure.subplots_adjust(left=0.08, right=0.98, top=0.95, bottom=0.07, hspace=0.001)

        self.canvas.draw()

    def plotxy(self, x, y):
        """
        Plot one variable against another
        """
        self.axes.clear()
        self.figure.clear()
        self.axes = self.figure.add_subplot(111)
        self.figure.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08)
        self.axes.plot(x.data, y.data)
        self.axes.set_xlabel(x.name)
        self.axes.set_ylabel(y.name)
        self.canvas.draw()

    def oplot(self, *args):
        """
        Make an overplot from multiple traces
        """
        ntraces = len(args)
        if len(args) == 1:
            self.plot(*args)

        self._clean_axes()

        self.axes = self.figure.add_subplot(1, 1, 1)
        for tracenum, trace in enumerate(args):
            try:
                for data in trace:
                    label = data.desc
                    if label == "":
                        label = data.name + " (" + data.units + ") " + data.source

                    time = data.time
                    if time is None:
                        if len(data.dim) != 1:
                            print(data.dim)
                            raise ValueError("Cannot plot '"+data.label+"' as it has too many dimensions")
                        time = data.dim[0].data

                    self.axes.plot(time, data.data, label=label)
                if tracenum == 0:
                    ylabel = data.desc
                    if ylabel == "":
                        ylabel = data.label
                        if data.units != "":
                            ylabel += " ("+data.units+") "
                    self.axes.set_ylabel(ylabel)
                    self.axes.set_xlabel(trace[0].dim[trace[0].order].label)
                self.axes.legend()
            except TypeError:
                #Trace not iterable
                time = trace.time.data
                xlabel = trace.dim[trace.order].label
                if time is None:
                    if len(trace.dim) != 1:
                        raise ValueError("Cannot plot '"+trace.label+"' as it has too many dimensions")
                    time = trace.dim[0].data
                    xlabel = trace.dim[0].label

                data = trace.data
                # Check array size
                size = len(time)
                if size > 10000:
                    fac = int(len(time) / 10000)
                    time = time[::fac]
                    data = data[::fac]
                    print("Warning: too many samples (%d). Down-sampling to %d points" % (size, len(time)))

                self.axes.plot(time, data)

                ylabel = trace.desc
                if ylabel == "":
                    ylabel = trace.label
                    if trace.units != "":
                        ylabel += " ("+trace.units+") "
                self.axes.set_ylabel(ylabel)

        self.canvas.draw()

    def mplot(self, *args):
        """
        Make a custom number of plots with a custom number of traces on each
        """
        # Create a dialog box to get the plot format
        def getFormat():
            parent = QDialog()
            title = "Plot Format"
            label = 'Enter number of traces in each plot separated by ", ":'
            dialog = QInputDialog.getText(parent, title, label)
            if dialog[1]:
                form = dialog[0].split(',')
                form = [int(i) for i in form]
                return form
            else:
                raise ValueError("No format entered")

        pltform = getFormat()
        nplots = len(pltform)
        ntraces = len(args)

        # Check validity of the plot format
        if ntraces != np.sum(pltform):
            raise ValueError("Number of traces does not equal sum of traces to be plotted in format")

        self._clean_axes()
        # Split data items up into correct plots according to plot format
        traces = []
        maxindx = 0
        for i in range(nplots):
            plotdat = []
            minindx = maxindx
            maxindx += pltform[i]
            for j in range(ntraces):
                if j >= minindx and j < maxindx:
                    plotdat.append(args[j][0])
            traces.append(plotdat)

        # Set axes
        for plotnum in range(nplots):
            if plotnum == 0:
                self.axes = self.figure.add_subplot(nplots, 1, plotnum+1)
                ax = self.axes
            else:
                self.axes = self.figure.add_subplot(nplots, 1, plotnum+1, sharex=ax)
                setp(self.axes.get_xticklabels(), visible=False)

            # Create OPlots for each subfigure
            try:
                for data in traces[plotnum]:
                    label = data.desc
                    if label == "":
                        label = data.name + " (" + data.units + ")"
                    label += " " + data.source

                    time = data.time
                    if time is None:
                        if len(data.dim) != 1:
                            raise ValueError("Cannot plot '"+data.name+"' as it has too many dimensions")
                        time = data.dim[0].data

                    self.axes.plot(time, data.data, label=label)

                # Y label from last plot
                ylabel = data.desc
                if ylabel == "":
                    ylabel = data.label
                    if data.units != "":
                        ylabel += " ("+data.units+")"
                self.axes.set_ylabel(ylabel)

                if plotnum == 0:
                    self.axes.set_xlabel(traces[0][0].dim[traces[0][0].order].label)

                self.axes.legend()

            except TypeError:
                # Traces[plotnum] not iterable
                time = traces[plotnum].time.data
                xlabel = traces[plotnum].dim[traces[plotnum.order]].label
                if time is None:
                    if len(traces[plotnum].dim) != 1:
                        raise ValueError("Cannot plot '"+traces[plotnum].label+"' as it has too many dimensions")
                    time = traces[plotnum].dim[0].data
                    xlabel = traces[plotnum].dim[0].label

                data = traces[plotnum].data
                if size > 10000:
                    fac = int(len(time) / 10000)
                    time = time[::fac]
                    data = data[::fac]
                    print("Warning: too many samples (%d). Down-sampling to %d points" % (size, len(time)))

                self.axes.plot(time, data)

                ylabel = traces[plotnum].desc
                if ylabel == "":
                    ylabel = traces[plotnum].label
                    if traces[plotnum].units != "":
                        ylabel += " ("+traces[plotnum].units+")"
                self.axes.set_ylabel(ylabel)
                if plotnum == 0:
                    self.axes.set_xlabel(xlabel)

        self.figure.subplots_adjust(left=0.08, right=0.98, top=0.95, bottom=0.07, hspace=0.001)
        self.canvas.draw()

    def zplot(self, *args):
        """
        Plot traces and a zoomed in section on subplots
        """

        def connect_bbox(bbox1, bbox2,
                         loc1a, loc2a, loc1b, loc2b,
                         prop_lines, prop_patches=None):
            """
            Connect two bounding boxes
            """
            if prop_patches is None:
                prop_patches = prop_lines.copy()
                prop_patches["alpha"] = prop_patches.get("alpha", 1)*0.2

            c1 = BboxConnector(bbox1, bbox2, loc1=loc1a, loc2=loc2a, **prop_lines)
            c1.set_clip_on(False)
            c2 = BboxConnector(bbox1, bbox2, loc1=loc1b, loc2=loc2b, **prop_lines)
            c2.set_clip_on(False)

            bbox_patch1 = BboxPatch(bbox1, **prop_patches)
            bbox_patch2 = BboxPatch(bbox2, **prop_patches)

            patch = BboxConnectorPatch(bbox1, bbox2,
                                       loc1a=loc1a, loc2a=loc2a, loc1b=loc1b, loc2b=loc2b,
                                       **prop_patches)
            patch.set_clip_on(False)

            return c1, c2, bbox_patch1, bbox_patch2, patch

        def zoom_effect(ax1, ax2, **kwargs):
            """
            ax1 : the main axes
            ax1 : the zoomed axes

            Connect ax1 and ax2.  The xmin & xmax will be taken from the
            ax1.viewLim.
            """

            tt = ax1.transScale + (ax1.transLimits + ax2.transAxes)
            trans = blended_transform_factory(ax2.transData, tt)

            mybbox1 = ax1.bbox
            mybbox2 = TransformedBbox(ax1.viewLim, trans)

            prop_patches = kwargs.copy()
            prop_patches["ec"] = "none"
            prop_patches["alpha"] = 0.2

            c1, c2, bbox_patch1, bbox_patch2, patch = \
                connect_bbox(mybbox1, mybbox2,
                             loc1a=3, loc2a=2, loc1b=4, loc2b=1,
                             prop_lines=kwargs, prop_patches=prop_patches)

            ax1.add_patch(bbox_patch1)
            ax2.add_patch(bbox_patch2)
            ax2.add_patch(c1)
            ax2.add_patch(c2)
            ax2.add_patch(patch)

            return c1, c2, bbox_patch1, bbox_patch2, patch

        # if len(args) == 1:
        #     self.plot(*args)

        self._clean_axes()

        self.ax_top = self.figure.add_subplot(2, 1, 1)
        self.ax_bottom = self.figure.add_subplot(2, 1, 2)
        self.axes = [self.ax_top, self.ax_bottom]

        zoom_effect(self.ax_top, self.ax_bottom)
        self.start_cid = self.ax_bottom.figure.canvas.mpl_connect("button_press_event",
                                                                  self.start_zoom_region)
        self.end_cid = self.ax_bottom.figure.canvas.mpl_connect("button_release_event",
                                                                self.end_zoom_region)
        self.callback_id = [self.start_cid, self.end_cid]

        for tracenum, trace in enumerate(args):
            try:
                for data in trace:
                    label = data.desc
                    if label == "":
                        label = data.name + " (" + data.units + ") " + data.source

                    time = data.time
                    if time is None:
                        if len(data.dim) != 1:
                            print(data.dim)
                            raise ValueError("Cannot plot '"+data.label+"' as it has too many dimensions")
                        time = data.dim[0].data

                    self.ax_top.plot(time, data.data, label=label)
                    self.ax_bottom.plot(time, data.data, label=label)
                if tracenum == 0:
                    self.ax_bottom.set_xlabel(trace[0].dim[trace[0].order].label)
                    ylabel = data.desc
                    if ylabel == "":
                        ylabel = data.label
                        if data.units != "":
                            ylabel += " ("+data.units+") "
                    self.ax_bottom.set_ylabel(ylabel)
                    self.ax_top.set_ylabel(ylabel)
                self.ax_bottom.legend()
            except TypeError:
                # Trace not iterable
                time = trace.time.data
                xlabel = trace.dim[trace.order].label
                if time is None:
                    if len(trace.dim) != 1:
                        raise ValueError("Cannot plot '"+trace.label+"' as it has too many dimensions")
                    time = trace.dim[0].data
                    xlabel = trace.dim[0].label
                self.ax_bottom.set_xlabel(xlabel)

                data = trace.data
                # Check array size
                size = len(time)
                if size > 10000:
                    fac = int(len(time) / 10000)
                    time = time[::fac]
                    data = data[::fac]
                    print("Warning: too many samples (%d). Down-sampling to %d points" % (size, len(time)))

                self.ax_top.plot(time, data)
                self.ax_bottom.plot(time, data)

                ylabel = trace.desc
                if ylabel == "":
                    ylabel = trace.label
                    if trace.units != "":
                        ylabel += " ("+trace.units+") "
                self.ax_top.set_ylabel(ylabel)
                self.ax_bottom.set_ylabel(ylabel)

        self.figure.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08)
        self.canvas.draw()

    def start_zoom_region(self, event):
        """
        Callback for starting the zoomed region
        """
        if event.inaxes != self.ax_bottom:
            return
        _, right = self.ax_top.get_xlim()
        self.ax_top.set_xlim(event.xdata, right)
        self.ax_bottom.figure.canvas.draw()
        self.drag_cid = self.ax_bottom.figure.canvas.mpl_connect("motion_notify_event",
                                                                 self.middle_zoom_region)

    def middle_zoom_region(self, event):
        """
        Callback for dragging the zoomed region
        """
        if event.inaxes != self.ax_bottom:
            return
        left, _ = self.ax_top.get_xlim()
        self.ax_top.set_xlim(left, event.xdata)
        self.ax_bottom.figure.canvas.draw()

    def end_zoom_region(self, event):
        """
        Callback for ending the zoomed region
        """
        self.figure.canvas.mpl_disconnect(self.drag_cid)

    def contour(self, item, levels=10):
        """
        Contour plot of item

        item - Item to plot
        levels - Number of contours to plot [10]
        """

        self._clean_axes()
        self.axes = self.figure.add_subplot(111)
        self.figure.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08)

        if not (len(item.dim) == 2 or
                len(item.data) == 2 or
                len(item.data.shape) == 2):
            raise ValueError("Data must be two dimensional")
            return

        xdim = item.dim[item.order]
        ydim = item.dim[item.order+1]

        xaxis = xdim.data
        yaxis = ydim.data
        zaxis = item.data
        self.axes.contour(xaxis, yaxis, zaxis, levels)

        xlabel = xdim.label
        if xlabel == "":
            xlabel = xdim.name
            if xdim.units != "":
                xlabel += " ({})".format(xdim.units)

        self.axes.set_xlabel(xlabel)

        ylabel = ydim.label
        if ylabel == "":
            ylabel = ydim.name
            if ydim.units != "":
                ylabel += " ({})".format(ydim.units)

        self.axes.set_ylabel(ylabel)

        title = item.name
        if title == "":
            title = item.title
        self.axes.set_title(title)

        self.canvas.draw()

    def contourf(self, item, levels=10):
        """
        Filled contour plot of item

        item - Item to plot
        levels - Number of contours to plot [10]
        """

        self._clean_axes()
        self.axes = self.figure.add_subplot(111)
        self.figure.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08)

        if not (len(item.dim) == 2 or
                len(item.data) == 2 or
                len(item.data.shape) == 2):
            raise ValueError("Data must be two dimensional")
            return

        xdim = item.dim[item.order]
        ydim = item.dim[item.order+1]

        xaxis = xdim.data
        yaxis = ydim.data
        zaxis = item.data
        self.axes.contourf(xaxis, yaxis, zaxis, levels)

        xlabel = xdim.label
        if xlabel == "":
            xlabel = xdim.name
            if xdim.units != "":
                xlabel += " ({})".format(xdim.units)

        self.axes.set_xlabel(xlabel)

        ylabel = ydim.label
        if ylabel == "":
            ylabel = ydim.name
            if ydim.units != "":
                ylabel += " ({})".format(ydim.units)

        self.axes.set_ylabel(ylabel)

        title = item.name
        if title == "":
            title = item.title
        self.axes.set_title(title)

        self.canvas.draw()

    def clearFig(self):
        """
        Reset the plot widget
        """
        self._clean_axes()
        self.axes = self.figure.add_subplot(111)
        self.figure.subplots_adjust(left=0.08, right=0.98, top=0.95, bottom=0.07)
        self.canvas.draw()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pyxpad_main.ui'
#
# Created: Fri Oct  7 15:25:29 2016
#      by: pyside-uic 0.2.15 running on PySide 1.2.4
#
# WARNING! All changes made in this file will be lost!

from PySide import QtCore, QtGui

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(803, 659)
        self.centralwidget = QtGui.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout_2 = QtGui.QGridLayout(self.centralwidget)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.tabWidget = QtGui.QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")
        self.sourceTab = QtGui.QWidget()
        self.sourceTab.setObjectName("sourceTab")
        self.gridLayout_3 = QtGui.QGridLayout(self.sourceTab)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.gridLayout = QtGui.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.sourceDescription = QtGui.QCheckBox(self.sourceTab)
        self.sourceDescription.setObjectName("sourceDescription")
        self.gridLayout.addWidget(self.sourceDescription, 0, 6, 1, 1)
        self.shotLabel = QtGui.QLabel(self.sourceTab)
        self.shotLabel.setObjectName("shotLabel")
        self.gridLayout.addWidget(self.shotLabel, 0, 0, 1, 1)
        self.tracePattern = QtGui.QLineEdit(self.sourceTab)
        self.tracePattern.setObjectName("tracePattern")
        self.gridLayout.addWidget(self.tracePattern, 0, 5, 1, 1)
        self.shotInput = QtGui.QLineEdit(self.sourceTab)
        self.shotInput.setObjectName("shotInput")
        self.gridLayout.addWidget(self.shotInput, 0, 1, 1, 1)
        self.readDataButton = QtGui.QPushButton(self.sourceTab)
        self.readDataButton.setObjectName("readDataButton")
        self.gridLayout.addWidget(self.readDataButton, 0, 2, 1, 1)
        self.traceLabel = QtGui.QLabel(self.sourceTab)
        self.traceLabel.setObjectName("traceLabel")
        self.gridLayout.addWidget(self.traceLabel, 0, 4, 1, 1)
        self.lastShotButton = QtGui.QPushButton(self.sourceTab)
        self.lastShotButton.setObjectName("lastShotButton")
        self.gridLayout.addWidget(self.lastShotButton, 0, 3, 1, 1)
        self.gridLayout_3.addLayout(self.gridLayout, 0, 0, 1, 1)
        self.splitter = QtGui.QSplitter(self.sourceTab)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.treeView = QtGui.QTreeWidget(self.splitter)
        self.treeView.setColumnCount(1)
        self.treeView.setObjectName("treeView")
        self.treeView.headerItem().setText(0, "Source")
        self.treeView.header().setVisible(False)
        self.treeView.header().setDefaultSectionSize(200)
        self.treeView.header().setStretchLastSection(True)
        self.sourceTable = QtGui.QTableWidget(self.splitter)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sourceTable.sizePolicy().hasHeightForWidth())
        self.sourceTable.setSizePolicy(sizePolicy)
        self.sourceTable.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        self.sourceTable.setAlternatingRowColors(True)
        self.sourceTable.setSelectionMode(QtGui.QAbstractItemView.MultiSelection)
        self.sourceTable.setShowGrid(False)
        self.sourceTable.setColumnCount(1)
        self.sourceTable.setObjectName("sourceTable")
        self.sourceTable.setColumnCount(1)
        self.sourceTable.setRowCount(0)
        self.sourceTable.horizontalHeader().setVisible(False)
        self.sourceTable.horizontalHeader().setDefaultSectionSize(200)
        self.sourceTable.horizontalHeader().setStretchLastSection(False)
        self.sourceTable.verticalHeader().setVisible(False)
        self.sourceTable.verticalHeader().setDefaultSectionSize(20)
        self.gridLayout_3.addWidget(self.splitter, 1, 0, 1, 1)
        self.tabWidget.addTab(self.sourceTab, "")
        self.dataTab = QtGui.QWidget()
        self.dataTab.setObjectName("dataTab")
        self.gridLayout_5 = QtGui.QGridLayout(self.dataTab)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.splitter_2 = QtGui.QSplitter(self.dataTab)
        self.splitter_2.setOrientation(QtCore.Qt.Vertical)
        self.splitter_2.setObjectName("splitter_2")
        self.dataTable = QtGui.QTableWidget(self.splitter_2)
        self.dataTable.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.dataTable.setAlternatingRowColors(True)
        self.dataTable.setSelectionMode(QtGui.QAbstractItemView.MultiSelection)
        self.dataTable.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.dataTable.setShowGrid(False)
        self.dataTable.setGridStyle(QtCore.Qt.SolidLine)
        self.dataTable.setObjectName("dataTable")
        self.dataTable.setColumnCount(4)
        self.dataTable.setRowCount(0)
        item = QtGui.QTableWidgetItem()
        self.dataTable.setHorizontalHeaderItem(0, item)
        item = QtGui.QTableWidgetItem()
        self.dataTable.setHorizontalHeaderItem(1, item)
        item = QtGui.QTableWidgetItem()
        self.dataTable.setHorizontalHeaderItem(2, item)
        item = QtGui.QTableWidgetItem()
        self.dataTable.setHorizontalHeaderItem(3, item)
        self.dataTable.horizontalHeader().setStretchLastSection(True)
        self.dataTable.verticalHeader().setDefaultSectionSize(20)
        self.textOutput = QtGui.QTextEdit(self.splitter_2)
        self.textOutput.setReadOnly(True)
        self.textOutput.setObjectName("textOutput")
        self.layoutWidget = QtGui.QWidget(self.splitter_2)
        self.layoutWidget.setObjectName("layoutWidget")
        self.gridLayout_4 = QtGui.QGridLayout(self.layoutWidget)
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.label_3 = QtGui.QLabel(self.layoutWidget)
        self.label_3.setObjectName("label_3")
        self.gridLayout_4.addWidget(self.label_3, 0, 0, 1, 1)
        self.commandInput = ConsoleWidget(self.layoutWidget)
        self.commandInput.setObjectName("commandInput")
        self.gridLayout_4.addWidget(self.commandInput, 0, 1, 1, 1)
        self.commandButton = QtGui.QPushButton(self.layoutWidget)
        self.commandButton.setObjectName("commandButton")
        self.gridLayout_4.addWidget(self.commandButton, 0, 2, 1, 1)
        self.gridLayout_5.addWidget(self.splitter_2, 0, 0, 1, 1)
        self.tabWidget.addTab(self.dataTab, "")
        self.plotTab = QtGui.QWidget()
        self.plotTab.setObjectName("plotTab")
        self.tabWidget.addTab(self.plotTab, "")
        self.gridLayout_2.addWidget(self.tabWidget, 0, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtGui.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 803, 19))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtGui.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuAdd_source = QtGui.QMenu(self.menuFile)
        self.menuAdd_source.setObjectName("menuAdd_source")
        self.menuPlot = QtGui.QMenu(self.menubar)
        self.menuPlot.setObjectName("menuPlot")
        self.menuCommand = QtGui.QMenu(self.menubar)
        self.menuCommand.setObjectName("menuCommand")
        self.menuHelp = QtGui.QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtGui.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionLoadState = QtGui.QAction(MainWindow)
        self.actionLoadState.setObjectName("actionLoadState")
        self.actionSaveState = QtGui.QAction(MainWindow)
        self.actionSaveState.setObjectName("actionSaveState")
        self.actionExit = QtGui.QAction(MainWindow)
        self.actionExit.setObjectName("actionExit")
        self.actionXPAD_tree = QtGui.QAction(MainWindow)
        self.actionXPAD_tree.setObjectName("actionXPAD_tree")
        self.actionNetCDF_file = QtGui.QAction(MainWindow)
        self.actionNetCDF_file.setObjectName("actionNetCDF_file")
        self.actionPlot = QtGui.QAction(MainWindow)
        self.actionPlot.setObjectName("actionPlot")
        self.actionOPlot = QtGui.QAction(MainWindow)
        self.actionOPlot.setObjectName("actionOPlot")
        self.actionMPlot = QtGui.QAction(MainWindow)
        self.actionMPlot.setObjectName("actionMPlot")
        self.actionXYPlot = QtGui.QAction(MainWindow)
        self.actionXYPlot.setObjectName("actionXYPlot")
        self.actionZPlot = QtGui.QAction(MainWindow)
        self.actionZPlot.setObjectName("actionZPlot")
        self.actionBOUT_data = QtGui.QAction(MainWindow)
        self.actionBOUT_data.setObjectName("actionBOUT_data")
        self.actionContour = QtGui.QAction(MainWindow)
        self.actionContour.setObjectName("actionContour")
        self.actionContour_filled = QtGui.QAction(MainWindow)
        self.actionContour_filled.setObjectName("actionContour_filled")
        self.actionClearFig = QtGui.QAction(MainWindow)
        self.actionClearFig.setObjectName("actionClearFig")
        self.actionWrite_ASCII = QtGui.QAction(MainWindow)
        self.actionWrite_ASCII.setObjectName("actionWrite_ASCII")
        self.actionAdd = QtGui.QAction(MainWindow)
        self.actionAdd.setObjectName("actionAdd")
        self.actionSubtract = QtGui.QAction(MainWindow)
        self.actionSubtract.setObjectName("actionSubtract")
        self.actionMultiply = QtGui.QAction(MainWindow)
        self.actionMultiply.setObjectName("actionMultiply")
        self.actionDivide = QtGui.QAction(MainWindow)
        self.actionDivide.setObjectName("actionDivide")
        self.actionChop = QtGui.QAction(MainWindow)
        self.actionChop.setObjectName("actionChop")
        self.actionIntegrate = QtGui.QAction(MainWindow)
        self.actionIntegrate.setObjectName("actionIntegrate")
        self.actionDf_dt = QtGui.QAction(MainWindow)
        self.actionDf_dt.setObjectName("actionDf_dt")
        self.actionSmooth = QtGui.QAction(MainWindow)
        self.actionSmooth.setObjectName("actionSmooth")
        self.actionLow_pass_filter = QtGui.QAction(MainWindow)
        self.actionLow_pass_filter.setObjectName("actionLow_pass_filter")
        self.actionHigh_pass_filter = QtGui.QAction(MainWindow)
        self.actionHigh_pass_filter.setObjectName("actionHigh_pass_filter")
        self.actionBand_pass_filter = QtGui.QAction(MainWindow)
        self.actionBand_pass_filter.setObjectName("actionBand_pass_filter")
        self.actionFFTP = QtGui.QAction(MainWindow)
        self.actionFFTP.setObjectName("actionFFTP")
        self.actionRunFFT = QtGui.QAction(MainWindow)
        self.actionRunFFT.setObjectName("actionRunFFT")
        self.actionReciprocal = QtGui.QAction(MainWindow)
        self.actionReciprocal.setObjectName("actionReciprocal")
        self.actionExponential = QtGui.QAction(MainWindow)
        self.actionExponential.setObjectName("actionExponential")
        self.actionAbsolute = QtGui.QAction(MainWindow)
        self.actionAbsolute.setObjectName("actionAbsolute")
        self.actionArctan = QtGui.QAction(MainWindow)
        self.actionArctan.setObjectName("actionArctan")
        self.actionNlog = QtGui.QAction(MainWindow)
        self.actionNlog.setObjectName("actionNlog")
        self.actionNorm = QtGui.QAction(MainWindow)
        self.actionNorm.setObjectName("actionNorm")
        self.actionInvert = QtGui.QAction(MainWindow)
        self.actionInvert.setObjectName("actionInvert")
        self.actionAddCon = QtGui.QAction(MainWindow)
        self.actionAddCon.setObjectName("actionAddCon")
        self.actionSubCon = QtGui.QAction(MainWindow)
        self.actionSubCon.setObjectName("actionSubCon")
        self.actionMulCon = QtGui.QAction(MainWindow)
        self.actionMulCon.setObjectName("actionMulCon")
        self.actionDivCon = QtGui.QAction(MainWindow)
        self.actionDivCon.setObjectName("actionDivCon")
        self.actionPowCon = QtGui.QAction(MainWindow)
        self.actionPowCon.setObjectName("actionPowCon")
        self.actionChangeName = QtGui.QAction(MainWindow)
        self.actionChangeName.setObjectName("actionChangeName")
        self.actionChangeUnits = QtGui.QAction(MainWindow)
        self.actionChangeUnits.setObjectName("actionChangeUnits")
        self.actionClip = QtGui.QAction(MainWindow)
        self.actionClip.setObjectName("actionClip")
        self.actionStats = QtGui.QAction(MainWindow)
        self.actionStats.setObjectName("actionStats")
        self.actionTimeOff = QtGui.QAction(MainWindow)
        self.actionTimeOff.setObjectName("actionTimeOff")
        self.actionAbout = QtGui.QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionDeleteTrace = QtGui.QAction(MainWindow)
        self.actionDeleteTrace.setObjectName("actionDeleteTrace")
        self.menuAdd_source.addAction(self.actionNetCDF_file)
        self.menuAdd_source.addAction(self.actionXPAD_tree)
        self.menuAdd_source.addAction(self.actionBOUT_data)
        self.menuFile.addAction(self.menuAdd_source.menuAction())
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionLoadState)
        self.menuFile.addAction(self.actionSaveState)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionWrite_ASCII)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionExit)
        self.menuPlot.addAction(self.actionPlot)
        self.menuPlot.addAction(self.actionOPlot)
        self.menuPlot.addAction(self.actionMPlot)
        self.menuPlot.addAction(self.actionZPlot)
        self.menuPlot.addAction(self.actionXYPlot)
        self.menuPlot.addAction(self.actionContour)
        self.menuPlot.addAction(self.actionContour_filled)
        self.menuPlot.addAction(self.actionClearFig)
        self.menuCommand.addAction(self.actionDeleteTrace)
        self.menuCommand.addSeparator()
        self.menuCommand.addAction(self.actionChop)
        self.menuCommand.addSeparator()
        self.menuCommand.addAction(self.actionIntegrate)
        self.menuCommand.addAction(self.actionDf_dt)
        self.menuCommand.addAction(self.actionSmooth)
        self.menuCommand.addAction(self.actionBand_pass_filter)
        self.menuCommand.addSeparator()
        self.menuCommand.addAction(self.actionAdd)
        self.menuCommand.addAction(self.actionSubtract)
        self.menuCommand.addAction(self.actionMultiply)
        self.menuCommand.addAction(self.actionDivide)
        self.menuCommand.addSeparator()
        self.menuCommand.addAction(self.actionFFTP)
        self.menuCommand.addAction(self.actionRunFFT)
        self.menuCommand.addSeparator()
        self.menuCommand.addAction(self.actionReciprocal)
        self.menuCommand.addAction(self.actionExponential)
        self.menuCommand.addAction(self.actionAbsolute)
        self.menuCommand.addAction(self.actionArctan)
        self.menuCommand.addAction(self.actionNlog)
        self.menuCommand.addAction(self.actionNorm)
        self.menuCommand.addAction(self.actionInvert)
        self.menuCommand.addAction(self.actionAddCon)
        self.menuCommand.addAction(self.actionSubCon)
        self.menuCommand.addAction(self.actionMulCon)
        self.menuCommand.addAction(self.actionDivCon)
        self.menuCommand.addAction(self.actionPowCon)
        self.menuCommand.addAction(self.actionChangeName)
        self.menuCommand.addAction(self.actionChangeUnits)
        self.menuCommand.addAction(self.actionClip)
        self.menuCommand.addAction(self.actionStats)
        self.menuCommand.addAction(self.actionTimeOff)
        self.menuHelp.addAction(self.actionAbout)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuPlot.menuAction())
        self.menubar.addAction(self.menuCommand.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QtGui.QApplication.translate("MainWindow", "PyXPad", None, QtGui.QApplication.UnicodeUTF8))
        self.sourceDescription.setText(QtGui.QApplication.translate("MainWindow", "Description", None, QtGui.QApplication.UnicodeUTF8))
        self.shotLabel.setText(QtGui.QApplication.translate("MainWindow", "Shot:", None, QtGui.QApplication.UnicodeUTF8))
        self.readDataButton.setText(QtGui.QApplication.translate("MainWindow", "&Read", None, QtGui.QApplication.UnicodeUTF8))
        self.traceLabel.setText(QtGui.QApplication.translate("MainWindow", "Trace:", None, QtGui.QApplication.UnicodeUTF8))
        self.lastShotButton.setToolTip(QtGui.QApplication.translate("MainWindow", "Get last shot number", None, QtGui.QApplication.UnicodeUTF8))
        self.lastShotButton.setText(QtGui.QApplication.translate("MainWindow", "&Last shot", None, QtGui.QApplication.UnicodeUTF8))
        self.sourceTable.setToolTip(QtGui.QApplication.translate("MainWindow", "List of available signals", None, QtGui.QApplication.UnicodeUTF8))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.sourceTab), QtGui.QApplication.translate("MainWindow", "&Sources", None, QtGui.QApplication.UnicodeUTF8))
        self.dataTable.horizontalHeaderItem(0).setText(QtGui.QApplication.translate("MainWindow", "Name", None, QtGui.QApplication.UnicodeUTF8))
        self.dataTable.horizontalHeaderItem(1).setText(QtGui.QApplication.translate("MainWindow", "Source", None, QtGui.QApplication.UnicodeUTF8))
        self.dataTable.horizontalHeaderItem(2).setText(QtGui.QApplication.translate("MainWindow", "Trace", None, QtGui.QApplication.UnicodeUTF8))
        self.dataTable.horizontalHeaderItem(3).setText(QtGui.QApplication.translate("MainWindow", "Comments", None, QtGui.QApplication.UnicodeUTF8))
        self.label_3.setText(QtGui.QApplication.translate("MainWindow", "Command:", None, QtGui.QApplication.UnicodeUTF8))
        self.commandButton.setText(QtGui.QApplication.translate("MainWindow", "Run", None, QtGui.QApplication.UnicodeUTF8))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.dataTab), QtGui.QApplication.translate("MainWindow", "&Data", None, QtGui.QApplication.UnicodeUTF8))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.plotTab), QtGui.QApplication.translate("MainWindow", "&Plot", None, QtGui.QApplication.UnicodeUTF8))
        self.menuFile.setTitle(QtGui.QApplication.translate("MainWindow", "&File", None, QtGui.QApplication.UnicodeUTF8))
        self.menuAdd_source.setTitle(QtGui.QApplication.translate("MainWindow", "&Add source", None, QtGui.QApplication.UnicodeUTF8))
        self.menuPlot.setTitle(QtGui.QApplication.translate("MainWindow", "&Graphics", None, QtGui.QApplication.UnicodeUTF8))
        self.menuCommand.setTitle(QtGui.QApplication.translate("MainWindow", "&Command", None, QtGui.QApplication.UnicodeUTF8))
        self.menuHelp.setTitle(QtGui.QApplication.translate("MainWindow", "&Help", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLoadState.setText(QtGui.QApplication.translate("MainWindow", "&Load state", None, QtGui.QApplication.UnicodeUTF8))
        self.actionSaveState.setText(QtGui.QApplication.translate("MainWindow", "&Save state", None, QtGui.QApplication.UnicodeUTF8))
        self.actionExit.setText(QtGui.QApplication.translate("MainWindow", "E&xit", None, QtGui.QApplication.UnicodeUTF8))
        self.actionExit.setToolTip(QtGui.QApplication.translate("MainWindow", "Exit pyXpad", None, QtGui.QApplication.UnicodeUTF8))
        self.actionExit.setShortcut(QtGui.QApplication.translate("MainWindow", "Ctrl+Q", None, QtGui.QApplication.UnicodeUTF8))
        self.actionXPAD_tree.setText(QtGui.QApplication.translate("MainWindow", "XPAD tree", None, QtGui.QApplication.UnicodeUTF8))
        self.actionXPAD_tree.setToolTip(QtGui.QApplication.translate("MainWindow", "Load a tree of XPAD items", None, QtGui.QApplication.UnicodeUTF8))
        self.actionNetCDF_file.setText(QtGui.QApplication.translate("MainWindow", "NetCDF file", None, QtGui.QApplication.UnicodeUTF8))
        self.actionPlot.setText(QtGui.QApplication.translate("MainWindow", "&Plot", None, QtGui.QApplication.UnicodeUTF8))
        self.actionOPlot.setText(QtGui.QApplication.translate("MainWindow", "&OPlot", None, QtGui.QApplication.UnicodeUTF8))
        self.actionMPlot.setText(QtGui.QApplication.translate("MainWindow", "&MPlot", None, QtGui.QApplication.UnicodeUTF8))
        self.actionXYPlot.setText(QtGui.QApplication.translate("MainWindow", "&XYPlot", None, QtGui.QApplication.UnicodeUTF8))
        self.actionZPlot.setText(QtGui.QApplication.translate("MainWindow", "&ZPlot", None, QtGui.QApplication.UnicodeUTF8))
        self.actionBOUT_data.setText(QtGui.QApplication.translate("MainWindow", "BOUT++ data", None, QtGui.QApplication.UnicodeUTF8))
        self.actionBOUT_data.setToolTip(QtGui.QApplication.translate("MainWindow", "Read BOUT++ data directory", None, QtGui.QApplication.UnicodeUTF8))
        self.actionBOUT_data.setShortcut(QtGui.QApplication.translate("MainWindow", "Ctrl+S", None, QtGui.QApplication.UnicodeUTF8))
        self.actionContour.setText(QtGui.QApplication.translate("MainWindow", "&Contour", None, QtGui.QApplication.UnicodeUTF8))
        self.actionContour_filled.setText(QtGui.QApplication.translate("MainWindow", "Contour &filled", None, QtGui.QApplication.UnicodeUTF8))
        self.actionClearFig.setText(QtGui.QApplication.translate("MainWindow", "C&lear Figure", None, QtGui.QApplication.UnicodeUTF8))
        self.actionWrite_ASCII.setText(QtGui.QApplication.translate("MainWindow", "&Write ASCII", None, QtGui.QApplication.UnicodeUTF8))
        self.actionAdd.setText(QtGui.QApplication.translate("MainWindow", "X+Y (Sum Channels)", None, QtGui.QApplication.UnicodeUTF8))
        self.actionSubtract.setText(QtGui.QApplication.translate("MainWindow", "X-Y", None, QtGui.QApplication.UnicodeUTF8))
        self.actionMultiply.setText(QtGui.QApplication.translate("MainWindow", "X*Y", None, QtGui.QApplication.UnicodeUTF8))
        self.actionDivide.setText(QtGui.QApplication.translate("MainWindow", "X/Y", None, QtGui.QApplication.UnicodeUTF8))
        self.actionChop.setText(QtGui.QApplication.translate("MainWindow", "Chop", None, QtGui.QApplication.UnicodeUTF8))
        self.actionIntegrate.setText(QtGui.QApplication.translate("MainWindow", "Integrate", None, QtGui.QApplication.UnicodeUTF8))
        self.actionDf_dt.setText(QtGui.QApplication.translate("MainWindow", "df/dt", None, QtGui.QApplication.UnicodeUTF8))
        self.actionSmooth.setText(QtGui.QApplication.translate("MainWindow", "Smooth", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLow_pass_filter.setText(QtGui.QApplication.translate("MainWindow", "Low pass filter", None, QtGui.QApplication.UnicodeUTF8))
        self.actionHigh_pass_filter.setText(QtGui.QApplication.translate("MainWindow", "High pass filter", None, QtGui.QApplication.UnicodeUTF8))
        self.actionBand_pass_filter.setText(QtGui.QApplication.translate("MainWindow", "Band pass filter", None, QtGui.QApplication.UnicodeUTF8))
        self.actionFFTP.setText(QtGui.QApplication.translate("MainWindow", "FFTP", None, QtGui.QApplication.UnicodeUTF8))
        self.actionRunFFT.setText(QtGui.QApplication.translate("MainWindow", "Running FFT", None, QtGui.QApplication.UnicodeUTF8))
        self.actionReciprocal.setText(QtGui.QApplication.translate("MainWindow", "1/X", None, QtGui.QApplication.UnicodeUTF8))
        self.actionExponential.setText(QtGui.QApplication.translate("MainWindow", "exp", None, QtGui.QApplication.UnicodeUTF8))
        self.actionAbsolute.setText(QtGui.QApplication.translate("MainWindow", "abs", None, QtGui.QApplication.UnicodeUTF8))
        self.actionArctan.setText(QtGui.QApplication.translate("MainWindow", "arctan", None, QtGui.QApplication.UnicodeUTF8))
        self.actionNlog.setText(QtGui.QApplication.translate("MainWindow", "ln", None, QtGui.QApplication.UnicodeUTF8))
        self.actionNorm.setText(QtGui.QApplication.translate("MainWindow", "Normalise", None, QtGui.QApplication.UnicodeUTF8))
        self.actionInvert.setText(QtGui.QApplication.translate("MainWindow", "Invert", None, QtGui.QApplication.UnicodeUTF8))
        self.actionAddCon.setText(QtGui.QApplication.translate("MainWindow", "X+C", None, QtGui.QApplication.UnicodeUTF8))
        self.actionSubCon.setText(QtGui.QApplication.translate("MainWindow", "X-C", None, QtGui.QApplication.UnicodeUTF8))
        self.actionMulCon.setText(QtGui.QApplication.translate("MainWindow", "X*C", None, QtGui.QApplication.UnicodeUTF8))
        self.actionDivCon.setText(QtGui.QApplication.translate("MainWindow", "X/C", None, QtGui.QApplication.UnicodeUTF8))
        self.actionPowCon.setText(QtGui.QApplication.translate("MainWindow", "X^C", None, QtGui.QApplication.UnicodeUTF8))
        self.actionChangeName.setText(QtGui.QApplication.translate("MainWindow", "Change Name", None, QtGui.QApplication.UnicodeUTF8))
        self.actionChangeUnits.setText(QtGui.QApplication.translate("MainWindow", "Change Units", None, QtGui.QApplication.UnicodeUTF8))
        self.actionClip.setText(QtGui.QApplication.translate("MainWindow", "Clip", None, QtGui.QApplication.UnicodeUTF8))
        self.actionStats.setText(QtGui.QApplication.translate("MainWindow", "Statistics", None, QtGui.QApplication.UnicodeUTF8))
        self.actionTimeOff.setText(QtGui.QApplication.translate("MainWindow", "Time Offset", None, QtGui.QApplication.UnicodeUTF8))
        self.actionAbout.setText(QtGui.QApplication.translate("MainWindow", "&About", None, QtGui.QApplication.UnicodeUTF8))
        self.actionDeleteTrace.setText(QtGui.QApplication.translate("MainWindow", "&Delete Trace", None, QtGui.QApplication.UnicodeUTF8))

from .console_widget import ConsoleWidget



"""
Creates small configuration dialogs from a dictionary of values
"""

# Author: Ben Dudson, Department of Physics, University of York
#         benjamin.dudson@york.ac.uk
#
# This file is part of PyXPad.
#
# PyXPad is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyXPad is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

from PySide.QtGui import QDialog, QGridLayout, QLineEdit, QLabel, QCheckBox, QDialogButtonBox
from PySide.QtCore import Qt


class ConfigDialog(QDialog):
    def __init__(self, settings, parent=None, description=None, pstvOnly = True):
        super(ConfigDialog, self).__init__(parent)

        self.setWindowTitle("Configure source")

        self.layout = QGridLayout(self)
        row = 0
        self.widgets = {}
        self.settings = settings

        # Settings should be a dictionary
        for name, val in settings.items():

            label = QLabel(self)
            label.setText(str(name))
            self.layout.addWidget(label, row, 0, 1, 1)

            # Check the type of each setting, and create widgets accordingly
            if isinstance(val, str):
                # A string of some kind
                widget = QLineEdit(self)
                widget.setText(val)
            elif isinstance(val, list):
                # A list of alternative values, first is selected
                print("List: ", name)
                continue
            elif isinstance(val, bool):
                widget = QCheckBox(self)
                if val:
                    widget.setCheckState(Qt.CheckState.Checked)
                else:
                    widget.setCheckState(Qt.CheckState.Unchecked)
            elif isinstance(val, int):
                widget = QLineEdit(self)
                widget.setInputMask("9000000")
                widget.setText(str(val).strip())
            elif isinstance(val, float):
                widget = QLineEdit(self)
                if pstvOnly:
                    widget.setInputMask("0.000")
                widget.setText(str(val).strip())
            else:
                print("Ignoring: " + name)
                continue
            widget.config = name
            self.widgets[name] = widget
            self.layout.addWidget(widget, row, 1, 1, 1)
            row += 1
        # Add OK and Cancel buttons
        buttonbox = QDialogButtonBox(QDialogButtonBox.Ok |
                                     QDialogButtonBox.Cancel)
        buttonbox.accepted.connect(self.getValues)
        buttonbox.rejected.connect(self.reject)
        self.layout.addWidget(buttonbox, row, 1, 2, 1)

    def getValues(self):
        # Loop through widgets to extract values
        for name, widget in self.widgets.items():
            val = self.settings[name]  # The old value
            if isinstance(val, str):
                self.settings[name] = widget.text()
            elif isinstance(val, bool):
                self.settings[name] = widget.isChecked()
            elif isinstance(val, int):
                self.settings[name] = int(widget.text())
            elif isinstance(val, float):
                self.settings[name] = float(widget.text())
        self.accept()



"""



"""

# Author: Ben Dudson, Department of Physics, University of York
#         benjamin.dudson@york.ac.uk
#
# This file is part of PyXPad.
#
# PyXPad is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyXPad is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

import os

from .pyxpad_utils import XPadDataItem, XPadDataDim

try:
    import pyidam as idam
    gotidam = True
except ImportError:
    try:
        import idam
        gotidam = True
    except ImportError:
        print("Warning: IDAM library not found. Cannot read data")
        gotidam = False


class XPadSource:
    def __init__(self, path, parent=None):

        # Convert path to string, strip NULL chars
        path = str(path)

        self.label = os.path.basename(os.path.normpath(path))
        self.dimensions = {}
        self.varNames = []
        self.variables = {}

        self.parent = parent

        # Define configuration dictionary
        if parent is None:
            self.config = {'Host': 'mast.fusion.org.uk',
                           'Port': 56565,
                           'verbose': True,
                           'debug': False}
        else:
            self.config = parent.config

        if os.path.isdir(path):
            # List directory
            ls = os.listdir(path)

            # Check if a name is supplied
            if 'title' in ls:
                # Read file to get the label
                f = open(os.path.join(path, 'title'), 'r')
                self.label = f.readline().strip()
                f.close()

            # Create a child for each subdirectory
            self.children = [XPadSource(os.path.join(path, name), parent=self)  # Create child
                             for name in ls
                             # For each directory which isn't hidden
                             if os.path.isdir(os.path.join(path, name)) and name[0] != '.']

            # Find items
            for name in ls:
                if os.path.isfile(os.path.join(path, name)) and (os.path.splitext(name)[1] == ".item"):
                    self.children.append(XPadSource(os.path.join(path, name), parent=self))
        else:
            # Given an item file to read
            f = open(path, 'r')
            self.label = f.readline().strip()  # First line is the label
            nitems = int((f.readline().split('$', 1))[0].strip())  # Number of items
            for i in range(nitems):
                line = f.readline()
                # Split at '$'
                s = line.split('$', 1)
                name = s[0].strip()
                if len(name) == 0:
                    continue
                try:
                    desc = s[1].strip()
                except:
                    desc = ""
                item = XPadDataItem()
                item.name = name
                item.label = item.desc = desc
                item.source = path
                self.variables[name] = item
                self.varNames.append(name)

            if parent is not None:
                parent.addVariables(self.variables)
            f.close()

    def addVariables(self, vardict):
        # Add to dictionary of variables and list of names
        for name, var in vardict.items():
            self.variables[name] = var
            self.varNames.append(name)

        if self.parent is not None:
            self.parent.addVariables(vardict)  # Variables go from children up to parent

    def read(self, name, shot):
        """ Read data from IDAM """
        if not gotidam:
            raise ImportError("No IDAM library available")
        try:
            if isinstance(name, unicode):
                name = name.encode('utf-8')
        except NameError:
            pass
        name = str(name).strip()
        shot = str(shot).strip()

        # Start client
        if not hasattr(self, "client"):
            # Set configuration
            idam.Client.server = self.config['Host']
            idam.Client.port = self.config['Port']
            self.client = idam.Client()

        # Read data
        data = self.client.get(name, shot)

        if hasattr(data, "dims") and not hasattr(data, "dim"):
            data.dim = data.dims

        #Give data a name
        try:
            data.name = name
            data.source = "Shot " + shot
        except AttributeError:
            # Probably IDAM has set something to be read-only property
            pass

        return XPadDataItem(data)

    def size(self, name):
        pass

    def __getstate__(self):
        # We need to remove the IDAM client in order to pickle
        # instances of this class
        if hasattr(self, "client"):
            state = self.__dict__.copy()
            del state['client']
            return state
        else:
            return self.__dict__



"""
Calculus on XPadDataItem objects

"""

from .pyxpad_utils import XPadDataItem
from numpy import zeros, cumsum


def integrate(item):
    """
    Integrate the given trace

    Inputs
    ------

    item  - an XPadDataItem object (or equivalent)

    Returns
    -------

    an XPadDataItem object

    """

    if len(item.dim) != 1:
        raise ValueError("Can only integrate 1D traces currently")

    # Create a result
    result = XPadDataItem()
    if item.name != "":
        result.name = "INTG( "+item.name+" )"
    result.source = item.source
    if item.label != "":
        result.label = "INTG( "+item.label+" )"
    if item.units != "":
        result.units = item.units+"*"+item.dim[0].units

    result.data = zeros(item.data.shape)

    time = item.dim[0].data

    result.data[1:] = cumsum(0.5*(time[1:]-time[0:-1])*(item.data[1:] + item.data[0:-1]))

    result.dim = item.dim
    result.order = item.order
    result.time = item.time

    return result


def differentiate(item):
    """
    Differentiates the given trace

    Inputs
    ------

    item  - an XPadDataItem object (or equivalent)

    Returns
    -------

    an XPadDataItem object

    """

    if len(item.dim) != 1:
        raise ValueError("Can only differentiate 1D traces")

    # Create a result
    result = XPadDataItem()
    if item.name != "":
        result.name = "Diff(" + item.name + ")"
    result.source = item.source
    if item.label != "":
        result.label = "Diff(" + item.label + ")"
    if item.units != "":
        result.units = item.units + item.dim[0].units + chr(0x207B) + chr(0x00B9)
    result.dim = item.dim
    result.order = item.order
    result.time = item.time

    time = item.dim[item.order].data

    result.data = zeros(len(item.data))

    for i in range(1, len(result.data)-1):
        result.data[i] = (item.data[i+1]-item.data[i-1])/(time[i+1]-time[i-1])

    result.data[-1] = result.data[-2]
    result.data[0] = result.data[1]

    return result



"""


"""

from boutdata.data import BoutData
from .pyxpad_utils import XPadDataItem, XPadDataDim

import os


class BoutDataSource:
    """

    Functions
      read( name, shot )   Input variable name (string)
                     Output is an XPadDataItem object or None

      size( name )   Returns variable size as a list. [] for scalar

    Attributes

      label         A short string to describe the source
      dimensions    A dictionary of XPadDataDim objects
      varNames      A list of variable names
      variables     A dictionary of XPadDataItem objects with empty data

    """
    def __init__(self, path, parent=None):
        self.label = path

        self.parent = parent
        self.children = []

        # List the directory
        ls = os.listdir(path)

        for name in ls:
            fullname = os.path.join(path, name)
            if os.path.isdir(fullname):
                try:
                    c = BoutDataSource(fullname, parent=self)
                    self.children.append(c)
                except:
                    print("No data in directory "+fullname)

        self.label = path
        self.dimensions = {}
        self.varNames = []
        self.variables = {}
        try:
            self.data = BoutData(path)

            self.varNames = self.data.varNames
            for i, v in enumerate(self.varNames):
                try:
                    # This is for Python 2.x. Python 3.x will raise a NameError
                    # since unicode -> str and str -> bytes
                    if isinstance(v, unicode):
                        v = v.encode('utf-8')
                    v = str(v).translate(None, '\0')
                except NameError:
                    if isinstance(v, str):
                        v = v.encode('utf-8')

                self.varNames[i] = v
        except:
            # No data in path. Check if any children
            if len(self.children) == 0:
                raise

    def read(self, name, shot):
        try:
            if isinstance(name, unicode):
                name = name.encode('utf-8')
        except NameError:
            pass
        name = str(name).translate(None, '\0')

        item = self.data.read(name)

        return XPadDataItem(item)



#!/usr/bin/env python3

import argparse
import os
from pyxpad import PyXPad
from PySide.QtGui import QApplication
import sys


def main():
    """
    Data visualisation and analysis tool in Python,
    intended to be familiar to users of XPAD.

    Primarily for IDAM data from the MAST tokamak experiment,
    but can be used to view NetCDF files currently.
    """
    # Add command line arguments
    parser = argparse.ArgumentParser(description=main.__doc__)
    parser.add_argument("-c", "--config", nargs=1, help="Config file to load",
                        default=None)
    parser.add_argument("-i", "--ignore-config", help="Ignore existing config files",
                        action="store_true", default=False)
    args = parser.parse_args()

    loadfile = args.config[0] if args.config is not None else None

    app = QApplication(sys.argv[0])
    window = PyXPad(loadfile=loadfile, ignoreconfig=args.ignore_config)
    window.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



"""
Fourier transform based methods on XPadDataItem objects
"""

from .pyxpad_utils import XPadDataItem, XPadDataDim

from numpy.fft import rfft, rfftfreq
from numpy import abs, arctan2, array, pi, zeros

def fftp(item):
    """
    Calculate amplitude and phase as a function of frequency

    Inputs
    ------

    item  - an XPadDataItem object

    Returns
    -------

    amplitude, phase pair of XPadDataItem objects

    """

    if len(item.dim) != 1:
        raise ValueError("fftp can only operate on 1D traces currently")

    # Calculate FFT
    data = rfft(item.data)*(1./len(item.data))

    # Create a dimension
    dim = XPadDataDim()

    dim.name = "Frequency"

    step = (item.dim[0].data[1] - item.dim[0].data[0])
    dim.data = rfftfreq(len(item.data), step)

    dim.units = "1/"+item.dim[0].units
    if item.dim[0].units in ["s", "S", "sec", "Sec", "SEC"]:
        dim.data /= 1000.
        dim.units = "kHz"

    # Calculate the amplitude
    amp = XPadDataItem()
    if item.name != "":
        amp.name = "AMP( "+item.name+" )"
    amp.source = item.source
    if item.label != "":
        amp.label = "AMP( "+item.label+" )"
    amp.units = item.units

    amp.data = abs(data)

    amp.dim = [dim]

    # Calculate the phase
    phase = XPadDataItem()
    if item.name != "":
        phase.name = "PHASE( "+item.name+" )"
    phase.source = item.source
    if item.label != "":
        phase.label = "PHASE( "+item.label+" )"
    phase.units = "Radians"

    phase.data = arctan2(data.real, data.imag)

    a = phase.data - 2*pi
    for i in range(1,len(phase.data)):
        if abs(phase.data[i-1] - a[i]) < 1.:
            phase.data[i] = a[i]

    a = phase.data + 2*pi
    for i in range(1,len(phase.data)):
        if abs(phase.data[i-1] - a[i]) < 1.:
            phase.data[i] = a[i]

    phase.dim = [dim]

    return amp, phase

def runfft(item, stride, width):
    """
    Performs a running Fourier transform on a data trace
    """

    # Check trace dimensions
    if len(item.dim) != 1:
        raise ValueError("runfft can only operate on 1D traces currently")

    time = item.dim[item.order]

    # Create dimensions
    time_dim = XPadDataDim(time)
    time_dim.data = []

    freq_dim = XPadDataDim()
    freq_dim.name = "Frequency"
    freq_dim.units = item.units + chr(0x207B) + chr(0x00B9)
    freq_dim.data = []
    if time_dim.units in ["s", "S", "sec", "Sec", "SEC"]:
        freq_dim.units = "kHz"

    # Assume time dimension is uniformly spaced
    # Explicit assumption of FFTW
    dt = time.data[1] - time.data[0]
    # Width and stride of window in index-space
    index_width = int(width / dt)
    index_stride = int(stride / dt)

    time_len = ((len(time.data) - index_width) // index_stride) + 1
    freq_len = (index_width // 2) + 1

    # Create result XPadDataItems for:
    # Amplitude
    amp = XPadDataItem()
    if item.name != "":
        amp.name = "runfft({}, stride={}, width={})".format(item.name, stride, width)
    amp.source = item.source
    if item.label != "":
        amp.label = "runfft({}, stride={}, width={})".format(item.name, stride, width)
    amp.units = item.units
    amp.dim = [freq_dim, time_dim]
    amp.order = -1
    amp.time = amp.dim[amp.order]
    amp.data = zeros((freq_len, time_len))

    window_freq = rfftfreq(index_width, dt)
    if time_dim.units in ["s", "S", "sec", "Sec", "SEC"]:
        window_freq /= 1000.
    freq_dim.data = window_freq

    for window_index, time_index in enumerate(range(0, len(time.data) - index_width, index_stride)):
        window_data = item.data[time_index:time_index + index_width]
        window_time = time.data[time_index:time_index + index_width]

        window_fft = rfft(window_data) / index_width
        amp.data[:, window_index]  = abs(window_fft)

        # Update time
        time_dim.data.append((window_time[0] + window_time[-1])/2.)

    return amp



from setuptools import setup

setup(name='pyxpad',
      version='0.2.0',
      description='Data visualisation and analysis tool for IDAM',
      author='Ben Dudson',
      author_email='benjamin.dudson@york.ac.uk',
      maintainer='Peter Hill',
      maintainer_email='peter.hill@york.ac.uk',
      url='https://github.com/ZedThree/pyxpad/',
      download_url='https://github.com/ZedThree/pyxpad/tarball/0.1.0',
      license='GPL',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: X11 Applications :: Qt',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Scientific/Engineering :: Visualization',
      ],
      packages=['pyxpad'],
      install_requires=['matplotlib', 'numpy', 'scipy', 'PySide', 'xdg'],
      entry_points={
          'gui_scripts': [
              'pyxpad = pyxpad.__main__:main',
          ],
      },
)



"""
A Python console custom widget for Qt Designer.

Origin: Merging of

https://qt.gitorious.org/pyside/pyside-examples/source/d4e4c7fdf71ab52083e49ffdea1b7daeff6c8d8d:examples/designer/plugins/widgets/pythonconsolewidget.py
http://www.nullege.com/codes/show/src@e@r@err-1.7.1@errbot@backends@graphic.py/32/PySide.QtGui.QPlainTextEdit

Modified 2014 Ben Dudson <benjamin.dudson@york.ac.uk>
   * Returns the command string without executing it
   * PySide rather than Qt calls


Copyright (C) 2006 David Boddie <david@boddie.org.uk>
Copyright (C) 2005-2006 Trolltech ASA. All rights reserved.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""

from PySide import QtCore, QtGui


class ConsoleWidget(QtGui.QLineEdit):
    """ConsoleWidget

    Provides a custom widget to accept Python expressions and emit output
    to other components via a custom signal.

    """

    # This signal is emitted when a command is entered
    commandEntered = QtCore.Signal(str)

    def __init__(self, parent=None):

        QtGui.QLineEdit.__init__(self, parent)

        self.history = []
        self.current = -1

        self.returnPressed.connect(self.execute)

    def keyReleaseEvent(self, event):

        if event.type() == QtCore.QEvent.KeyRelease:

            if event.key() == QtCore.Qt.Key_Up:
                current = max(0, self.current - 1)
                if 0 <= current < len(self.history):
                    self.setText(self.history[current])
                    self.current = current

                event.accept()

            elif event.key() == QtCore.Qt.Key_Down:
                current = min(len(self.history), self.current + 1)
                if 0 <= current < len(self.history):
                    self.setText(self.history[current])
                else:
                    self.clear()
                self.current = current

                event.accept()

    def execute(self):
        expression = self.text()

        # Clear the line edit, append the expression to the
        # history, and update the current command index.
        self.clear()
        self.history.append(expression)
        self.current = len(self.history)

        # Emit the text
        self.commandEntered.emit(expression)


if __name__ == "__main__":

    import sys

    def runCommand(expr):
        print(expr)

    app = QtGui.QApplication(sys.argv)
    widget = ConsoleWidget()
    widget.commandEntered.connect(runCommand)

    widget.show()
    sys.exit(app.exec_())
